package de.duehl.basics.io;

/*
 * Copyright 2019 Christian Dühl. All rights reserved.
 *
 * This program is free software. You can redistribute it and/or
 * modify it under the same terms as perl:
 *
 * general:  http://dev.perl.org/licenses/
 * GPL:      http://dev.perl.org/licenses/gpl1.html
 * artistic: http://dev.perl.org/licenses/artistic.html
 */

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.filechooser.FileSystemView;

import de.duehl.basics.collections.Stack;
import de.duehl.basics.io.data.FileWithModificationTime;
import de.duehl.basics.io.exceptions.DirectoryNotCreatedRuntimeException;
import de.duehl.basics.io.exceptions.FileNotFoundRuntimeException;
import de.duehl.basics.io.exceptions.IORuntimeException;
import de.duehl.basics.system.SystemTools;
import de.duehl.basics.text.Text;

/**
 * Diese Klasse enthält Hilfsmethoden rund um das Dateisystem.
 *
 * @version 1.01     2019-03-28
 * @author Christian Dühl
 */

public class FileHelper {

    public static final String NO_EXTENSION_FOUND = "OHNE_EXTENSION";

    /** Eine Instanz soll hiervon nicht angelegt werden! */
    private FileHelper() {}

    /**
     * Löscht ein Verzeichnis mit Inhalt.
     *
     * @param path
     *            Verzeichnis
     * @throws IORuntimeException
     *             Wenn sich eine Datei oder Verzeichnis nicht löschen lässt.
     */
    public static void deleteTree(File path) {
        if (!path.exists()) {
            return;
        }
        for (File file : path.listFiles()) {
            if (file.isDirectory()) {
                deleteTree(file);
            }
            else if (!file.delete()) {
                throw new IORuntimeException(file + " could not be deleted!");
            }
        }

        if (!path.delete()) {
            throw new IORuntimeException(path + " could not be deleted!");
        }
    }

    /**
     * Löscht ein Verzeichnis mit Inhalt.
     *
     * @param path
     *            Verzeichnis
     * @throws IORuntimeException
     *             Wenn sich eine Datei oder Verzeichnis nicht löschen lässt.
     */
    public static void deleteTree(String path) {
        deleteTree(new File(path));
    }

    /**
     * Löscht ein Verzeichnis mit Inhalt. Es ist nicht sicher, ob das Verzeichnis danach auch
     * wirklich gelöscht ist, aber er versucht jede Datei und jedes Verzeichnis zu löschen.
     *
     * @param path
     *            Verzeichnis
     */
    public static void deleteTreeIgnoreErrors(File path) {
        if (!path.exists()) {
            return;
        }
        for (File file : path.listFiles()) {
            if (file.isDirectory()) {
                deleteTreeIgnoreErrors(file);
            }
            else if (!file.delete()) {
                //throw new IORuntimeException(file + " could not be deleted!");
            }
        }

        if (!path.delete()) {
            //throw new IORuntimeException(path + " could not be deleted!");
        }
    }

    /**
     * Löscht ein Verzeichnis mit Inhalt. Es ist nicht sicher, ob das Verzeichnis danach auch
     * wirklich gelöscht ist, aber er versucht jede Datei und jedes Verzeichnis zu löschen.
     *
     * @param path
     *            Verzeichnis
     */
    public static void deleteTreeIgnoreErrors(String path) {
        deleteTreeIgnoreErrors(new File(path));
    }

    /**
     * Sucht alle Dateien mit einer bestimmten Endung in einem Verzeichnis und seinen
     * Unterverzeichnissen.
     *
     * Hierbei spielt die Groß-/Kleinschreibung bei den Extensionen wie auch den Dateinamen keine
     * Rolle.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param extensions
     *            Liste aller Endungen der gesuchten Dateien (etwa ".txt" oder ".java")
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findFiles(String start, String... extensions) {
        if (0 == extensions.length) {
            throw new IllegalArgumentException("Mindestens eine Endung muss angegebene werden.");
        }

        List<String> lowerExtensions = new ArrayList<>();
        for (String extension : extensions) {
            lowerExtensions.add(extension.toLowerCase());
        }

        List<String> files = new ArrayList<>();
        for (File file : new File(start).listFiles()) {
            String path = file.getPath();
            if (file.isDirectory()) {
                files.addAll(findFiles(path, extensions));
            }
            else {
                String lowerPath = path.toLowerCase();
                for (String lowerExtension : lowerExtensions) {
                    if (lowerPath.endsWith(lowerExtension)) {
                        files.add(path);
                    }
                }
            }
        }

        return files;
    }

    /**
     * Sucht alle Dateien mit einer bestimmten Endung in einem Verzeichnis und seinen
     * Unterverzeichnissen.
     *
     * Hierbei spielt die Groß-/Kleinschreibung bei den Extensionen wie auch den Dateinamen keine
     * Rolle.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param extensions
     *            Liste aller Endungen der gesuchten Dateien (etwa ".txt" oder ".java")
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findFilesNio2Old(String start, String... extensions) {
        if (0 == extensions.length) {
            throw new IllegalArgumentException("Mindestens eine Endung muss angegebene werden.");
        }

        List<String> lowerExtensions = new ArrayList<>();
        for (String extension : extensions) {
            lowerExtensions.add(extension.toLowerCase());
        }

        List<String> files = new ArrayList<>();
        for (String file : findAllFilesNio2(start)) {
            String lowerFile = file.toLowerCase();
            for (String lowerExtension : lowerExtensions) {
                if (lowerFile.endsWith(lowerExtension)) {
                    files.add(file);
                }
            }
        }

        return files;
    }

    /**
     * Sucht alle Dateien mit einer bestimmten Endung in einem Verzeichnis und seinen
     * Unterverzeichnissen.
     *
     * Hierbei spielt die Groß-/Kleinschreibung bei den Extensionen wie auch den Dateinamen keine
     * Rolle.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param extensions
     *            Liste aller Endungen der gesuchten Dateien (etwa ".txt" oder ".java")
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findFilesNio2(String start, String... extensions) {
        if (0 == extensions.length) {
            throw new IllegalArgumentException("Mindestens eine Endung muss angegebene werden.");
        }

        List<String> lowerExtensions = new ArrayList<>();
        for (String extension : extensions) {
            lowerExtensions.add(extension.toLowerCase());
        }

        List<String> files = new ArrayList<>();
        Path dir = Paths.get(start);
        try (DirectoryStream<Path> dirStream = Files.newDirectoryStream(dir)) {
            Iterator<Path> iterator = dirStream.iterator();
            while (iterator.hasNext()) {
                Path file = iterator.next();
                if (Files.isRegularFile(file)) {
                    String fileAsString = file.toString();
                    String lowerFile = fileAsString.toLowerCase();
                    for (String lowerExtension : lowerExtensions) {
                        if (lowerFile.endsWith(lowerExtension)) {
                            files.add(fileAsString);
                            break;
                        }
                    }
                }
                else if (Files.isDirectory(file)) {
                    files.addAll(findFilesNio2(file.toString(), extensions));
                }
            }
        }
        catch (IOException exception) {
            throw new IllegalStateException("Fehler bei der Suche nach Dateien mit den Endungen "
                    + extensions + " im Verzeichnis " + start, exception);
        }
        return files;
    }

    /**
     * Findet alle Dateien im angegebene Verzeichnis und seinen Unterverzeichnissen.
     *
     * @param start
     *            Verzeichnis das durchsucht wird.
     * @return Liste der gefundenen Dateien.
     */
    public static List<String> findAllFilesNio2(String start) {
        List<String> files = new ArrayList<>();
        Path dir = Paths.get(start);
        try (DirectoryStream<Path> dirStream = Files.newDirectoryStream(dir)) {
            Iterator<Path> iterator = dirStream.iterator();
            while (iterator.hasNext()) {
                Path file = iterator.next();
                if (Files.isRegularFile(file)) {
                    files.add(file.toString());
                }
                else if (Files.isDirectory(file)) {
                    files.addAll(findAllFilesNio2(file.toString()));
                }
            }
        }
        catch (IOException exception) {
            throw new IllegalStateException("Fehler bei der Suche nach Dateien im Verzeichnis "
                    + start, exception);
        }
        return files;
    }

    /**
     * Sucht alle Dateien in einem Verzeichnis ohne seine Unterverzeichnisse zu berücksichtigen.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findAllFilesInMainDirectory(String start) {
        List<String> files = new ArrayList<>();
        File startAsFile = new File(start);
        for (File file : startAsFile.listFiles()) {
            if (file.isFile()) {
                files.add(file.getPath());
            }
        }
        return files;
    }

    /**
     * Sucht alle Dateien in einem Verzeichnis ohne seine Unterverzeichnisse zu berücksichtigen.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findAllFilesInMainDirectoryNio2(String start) {
        List<String> files = new ArrayList<>();
        Path dir = Paths.get(start);
        try (DirectoryStream<Path> dirStream = Files.newDirectoryStream(dir)) {
            Iterator<Path> imageIter = dirStream.iterator();
            while (imageIter.hasNext()) {
                Path file = imageIter.next();
                if (Files.isRegularFile(file)) {
                    files.add(file.toString());
                }
            }
        }
        catch (IOException exception) {
            throw new IllegalStateException("Fehler bei der Suche nach Dateien im Verzeichnis "
                    + start, exception);
        }
        return files;
    }

    /**
     * Sucht alle Dateien in einem Verzeichnis ohne seine Unterverzeichnisse zu berücksichtigen.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @return Liste mit den gefundenen Dateien ohne Pfad.
     */
    public static List<String> findAllFilesInMainDirectoryWithoutPath(String start) {
        List<String> list = findAllFilesInMainDirectory(start);
        List<String> result = new ArrayList<>();
        for (String file : list) {
            String fileWithoutPath = getBareName(file);
            result.add(fileWithoutPath);
        }
        return result;
    }

    /**
     * Sucht alle Dateien mit einer bestimmten Endung in einem Verzeichnis ohne seine
     * Unterverzeichnisse zu berücksichtigen.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param extension
     *            Endung der gesuchten Dateien (etwa ".txt" oder ".java")
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findFilesInMainDirectory(String start, String extension) {
        List<String> files = new ArrayList<>();
        File startAsFile = new File(start);
        String lowerExtension = extension.toLowerCase();
        for (File file : startAsFile.listFiles()) {
            if (file.getPath().toLowerCase().endsWith(lowerExtension)) {
                files.add(file.getPath());
            }
        }
        return files;
    }

    /**
     * Sucht alle Dateien mit einer bestimmten Endung in einem Verzeichnis ohne seine
     * Unterverzeichnisse zu berücksichtigen.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param extension
     *            Endung der gesuchten Dateien (etwa ".txt" oder ".java"). Diese kann mit oder ohne
     *            führenden Punkt angegebene werden.
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findFilesInMainDirectoryNio2(String start, String extension) {
        List<String> files = new ArrayList<>();
        String ext = extension;
        if (ext.startsWith(".")) {
            ext = ext.substring(1);
        }
        Path dir = Paths.get(start);
        try (DirectoryStream<Path> dirStream = Files.newDirectoryStream(dir, "*." + ext)) {
            Iterator<Path> imageIter = dirStream.iterator();
            while (imageIter.hasNext()) {
                Path file = imageIter.next();
                if (Files.isRegularFile(file)) {
                    files.add(file.toString());
                }
            }
        }
        catch (IOException exception) {
            throw new IllegalStateException("Fehler bei der Suche nach Dateien mit der Endung "
                    + extension + " im Verzeichnis " + start, exception);
        }
        return files;
    }

    /**
     * Sucht alle Dateien mit einer bestimmten Endung in einem Verzeichnis ohne seine
     * Unterverzeichnisse zu berücksichtigen.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param extension
     *            Endung der gesuchten Dateien (etwa ".txt" oder ".java")
     * @return Liste mit den gefundenen Dateien ohne Pfad.
     */
    public static List<String> findFilesInMainDirectoryWithoutPath(String start, String extension) {
        List<String> list = findFilesInMainDirectory(start, extension);
        List<String> result = new ArrayList<>();
        for (String file : list) {
            String fileWithoutPath = getBareName(file);
            result.add(fileWithoutPath);
        }
        return result;
    }

    /**
     * Sucht alle Verzeichnisse in einem Verzeichnis ohne seine Unterverzeichnisse zu
     * berücksichtigen.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @return Liste mit den gefundenen Verzeichnissen.
     */
    public static List<String> findAllSubdirectoriesInMainDirectoryNio2(String start) {
        List<String> files = new ArrayList<>();
        Path dir = Paths.get(start);
        try (DirectoryStream<Path> dirStream = Files.newDirectoryStream(dir)) {
            Iterator<Path> imageIter = dirStream.iterator();
            while (imageIter.hasNext()) {
                Path file = imageIter.next();
                if (Files.isDirectory(file, LinkOption.NOFOLLOW_LINKS)) {
                    files.add(file.toString());
                }
            }
        }
        catch (IOException exception) {
            throw new IllegalStateException(
                    "Fehler bei der Suche nach Unterverzeichnissen im Verzeichnis " + start,
                    exception);
        }
        return files;
    }

    /**
     * Sucht alle direkten Unterverzeichnisse in einem Verzeichnis.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findSubdirectoriesInMainDirectory(String start) {
        List<String> files = new ArrayList<>();
        File startAsFile = new File(start);
        for (File file : startAsFile.listFiles()) {
            if (file.isDirectory()) {
                files.add(file.getPath());
            }
        }
        return files;
    }

    /**
     * Sucht alle direkten Unterverzeichnisse in einem Verzeichnis.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @return Liste mit den gefundenen Dateien ohne Pfad.
     */
    public static List<String> findSubdirectoriesInMainDirectoryWithoutPath(String start) {
        List<String> list = findSubdirectoriesInMainDirectory(start);
        List<String> result = new ArrayList<>();
        for (String file : list) {
            String fileWithoutPath = getBareName(file);
            result.add(fileWithoutPath);
        }
        return result;
    }

    /**
     * Sucht alle direkten Unterverzeichnisse in einem Verzeichnis.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param pattern
     *            Das Pattern muss exakt auf den Namen des Verzeichnisses (ohne Pfad) passen.
     * @return Liste mit den gefundenen Dateien.
     */
    public static List<String> findSubdirectoriesInMainDirectory(String start, Pattern pattern) {
        List<String> files = new ArrayList<>();
        File startAsFile = new File(start);
        for (File file : startAsFile.listFiles()) {
            if (file.isDirectory()) {
                String fileWithPath = file.getPath();
                String fileWithoutPath = getBareName(fileWithPath);
                Matcher matcher = pattern.matcher(fileWithoutPath);
                if (matcher.matches()) {
                    files.add(fileWithPath);
                }
            }
        }
        return files;
    }

    /**
     * Sucht alle direkten Unterverzeichnisse in einem Verzeichnis.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param pattern
     *            Das Pattern muss exakt auf den Namen des Verzeichnisses (ohne Pfad) passen.
     * @return Liste mit den gefundenen Dateien ohne Pfad.
     */
    public static List<String> findSubdirectoriesInMainDirectoryWithoutPath(String start,
            Pattern pattern) {
        List<String> list = findSubdirectoriesInMainDirectory(start, pattern);
        List<String> result = new ArrayList<>();
        for (String file : list) {
            String fileWithoutPath = getBareName(file);
            result.add(fileWithoutPath);
        }
        return result;
    }

    /**
     * Sucht alle Dateien in einem Verzeichnis ohne seine Unterverzeichnisse zu berücksichtigen.
     *
     * @param dir
     *            Verzeichnis, in dem die Suche beginnt.
     * @return Liste mit den gefundenen Dateien (mit Pfaden).
     */
    public static List<String> findAllNormalFilesInMainDirectory(String dir) {
        List<String> files = new ArrayList<>();
        File dirAsFile = new File(dir);
        for (File file : dirAsFile.listFiles()) {
            if (file.isFile()) {
                files.add(file.getPath());
            }
        }
        return files;
    }

    /**
     * Ermittelt, ob es mindestens eine Datei mit einer bestimmten Endung in einem Verzeichnis und
     * seinen Unterverzeichnissen gibt.
     *
     * @param start
     *            Verzeichnis, in dem die Suche beginnt.
     * @param extension
     *            Endung der gesuchten Dateien (etwa ".txt" oder ".java")
     * @return Liste mit den gefundenen Dateien.
     */
    public static boolean containsFiles(String start, String extension) {
        File startAsFile = new File(start);
        for (File file : startAsFile.listFiles()) {
            if (file.isDirectory()) {
                if (containsFiles(file.getPath(), extension)) {
                    return true;
                }
            }
            else {
                if (file.getPath().endsWith(extension)) {
                    // System.out.println("file " + file);
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Ermittelt die Position des letzten Slash oder Backslash in einem Dateinamen.
     *
     * @param filename
     *            Zu untersuchender Dateiname.
     * @return Letzte Position (beginnend bei 0) oder -1, wenn weder Slash noch Backslash in der
     *         Eingabe enthalten waren.
     */
    public static int lastSlash(String filename) {
        int lastSlash = filename.lastIndexOf("/");
        int lastBackslash = filename.lastIndexOf("\\");
        if (lastSlash > lastBackslash) {
            return lastSlash;
        }
        else {
            return lastBackslash;
        }
    }

    /**
     * Normalisiert alle Backslashes zu Slashes.
     *
     * @param path
     *            Zu normalisierender Pfad.
     * @return Normalisierter Pfad.
     */
    public static String normalizeToSlahes(String path) {
        return path.replace("\\", "/");
    }

    /**
     * Normalisiert alle Slashes zu Backslashes.
     *
     * @param path
     *            Zu normalisierender Pfad.
     * @return Normalisierter Pfad.
     */
    public static String normalizeToBackSlahes(String path) {
        return path.replace("/", "\\");
    }

    /**
     * Ermittelt den Teil des übergebenen Dateinamens vor und inklusive des letzten Slashs oder
     * Backslashs.
     *
     * @param filename
     *            Zu untersuchender Dateiname.
     * @return Teil des Dateinamens vor dem letzten Slash oder Backslash inklusive dieses Zeichens.
     */
    public static String getDirName(String filename) {
        int index = lastSlash(filename);
        if (index == filename.length() - 1) {
            return getDirName(filename.substring(0, filename.length() - 1));
        }
        else {
            return filename.substring(0, index + 1);
        }
    }

    /**
     * Ermittelt den Teil des übergebenen Dateinamens vor und inklusive des letzten Slashs oder
     * Backslashs.
     *
     * @param filename
     *            Zu untersuchender Dateiname.
     * @return Teil des Dateinamens vor dem letzten Slash oder Backslash inklusive dieses Zeichens.
     */
    public static String getDirNameWithoutLastSlash(String filename) {
        String dirNameWithSlash = getDirName(filename);
        return dirNameWithSlash.substring(0, dirNameWithSlash.length() - 1);
    }

    /**
     * Ermittelt den Teil des übergebenen Dateinamens nach dem letzten Slash oder Backslash.
     *
     * @param file
     *            Zu untersuchende Datei.
     * @return Teil des Dateinamens nach dem letzten Slash oder Backslash. Falls der String mit
     *         einem solchen Zeichen endet, wird ein leerer String zurückgegeben.
     */
    public static String getBareName(File file) {
        return getBareName(file.getPath());
    }

    /**
     * Ermittelt den Teil des übergebenen Dateinamens nach dem letzten Slash oder Backslash.
     *
     * @param filename
     *            Zu untersuchender Dateiname.
     * @return Teil des Dateinamens nach dem letzten Slash oder Backslash. Falls der String mit
     *         einem solchen Zeichen endet, wird ein leerer String zurückgegeben.
     */
    public static String getBareName(String filename) {
        int index = lastSlash(filename);
        String bareFileName;
        if (filename.length() > index + 1) {
            bareFileName = filename.substring(index + 1);
        }
        else {
            bareFileName = "";
        }
        return bareFileName;

        // alternativ: (new File(filename)).getName();
    }

    /**
     * Es wird der Pfad für ~/localDirectory zurückgegeben. Falls es nicht existiert, wird es
     * angelegt.
     *
     * @param subDirWithoutPath
     *            Name des lokalen Verzeichnisses (etwa ".Jarify").
     * @return Pfad für ~/localDirectory.
     * @throws FileNotFoundRuntimeException
     *             Falls das Home-Verzeichnis nicht existiert.
     * @throws DirectoryNotCreatedRuntimeException
     *             Falls ~/localDirectory nicht existiert und nicht angelegt werden kann.
     */
    public static String getAndPerhapsCreateDirectoryAtHome(String subDirWithoutPath) {
        String home = SystemTools.getHomeDirectory();
        return createSubdirectoryInExistingDirectoryIfNecessary(home, subDirWithoutPath);
    }

    /**
     * Es wird ein Unterverzeichnis in einem bestehenden Verzeichnis angelegt,
     * falls es noch nicht besteht.                                              <br/><br/>
     *
     * Im Beispiel mit dir = c:/temp und subDirWithoutPath = blubb wird das
     * Verzeichnis c:/temp/blubb angelegt, wenn es c:/temp gibt und
     * c:/temp/blubb noch nicht gibt.
     *
     * @param dir
     *            Name des bestehenden, übergeordneten Verzeichnisses (etwa
     *            "c:/temp").
     * @param subDirWithoutPath
     *            Name des Unterverzeichnisses (ohne Pfad), das darunter
     *            angelegt werden soll, etwa "blubb".
     * @return Verzeichnis, das angelegt wurde, falls es nicht schon existierte.
     * @throws FileNotFoundRuntimeException
     *             Falls das Home-Verzeichnis nicht existiert.
     * @throws DirectoryNotCreatedRuntimeException
     *             Falls ~/localDirectory nicht existiert und nicht angelegt
     *             werden kann.
     */
    public static String createSubdirectoryInExistingDirectoryIfNecessary(String dir,
            String subDirWithoutPath) {
        if (!isDirectory(dir)) {
            throw new FileNotFoundRuntimeException("Das Verzeichnis '" + dir + "' existiert nicht!");
        }

        String dirAtDir = concatPathes(dir, subDirWithoutPath);
        createDirectoryIfNotExists(dirAtDir);

        return dirAtDir;
    }

    /**
     * Legt ein Verzeichnis an.
     *
     * @param dirName
     *            Name des Verzeichnisses.
     * @return true genau dann, wenn das Verzeichnis angelegt wurde. false anderenfalls.
     */
    public static boolean createDirectory(String dirName) {
        File dir = new File(dirName);
        boolean success = dir.mkdir();
        return success;
    }

    /**
     * Legt ein Verzeichnis samt eventuell nich vorhandener Unterverzeichnisse an, falls sie nicht
     * vorhanden sind.
     *
     * @param dirName
     *            Name des Verzeichnisses.
     * @throws DirectoryNotCreatedRuntimeException
     *             Falls das Verzeichnis nicht angelegt werden konnte, obwohl es nicht existierte.
     */
    public static void createDirectoryWithMissingSubdirectoriesIfNotExists(String dirName) {
        Stack<String> missingSubdirectories = new Stack<>();

        String path = Text.removeTrailingSlash(dirName);
        while (!exists(path)) {
            String subDir = getBareName(path);
            path = getDirNameWithoutLastSlash(path);
            missingSubdirectories.push(subDir);
        }

        while (!missingSubdirectories.isEmpty()) {
            String subDir = missingSubdirectories.pop();
            path = concatPathes(path, subDir);
            createDirectory(path);
        }
    }

    /**
     * Legt ein Verzeichnis an, falls es nicht vorhanden ist.
     *
     * @param dirName
     *            Name des Verzeichnisses.
     * @throws DirectoryNotCreatedRuntimeException
     *             Falls das Verzeichnis nicht angelegt werden konnte, obwohl es nicht existierte.
     */
    public static void createDirectoryIfNotExists(String dirName) {
        if (!exists(dirName)) {
            boolean success = createDirectory(dirName);
            if (!success) {
                throw new DirectoryNotCreatedRuntimeException("Das Verzeichnis " + dirName
                        + " kann nicht angelegt werden.");
            }
        }
    }

    /**
     * Legt ein Verzeichnis an, das vorher nicht vorhanden sein darf.
     *
     * @param dirName
     *            Name des Verzeichnisses.
     * @throws DirectoryNotCreatedRuntimeException
     *             Falls das Verzeichnis nicht angelegt werden konnte, obwohl es nicht existierte.
     */
    public static void createNotExistingDirectory(String dirName) {
        if (exists(dirName)) {
            throw new RuntimeException("Das Verzeichnis '" + dirName
                    + "', das angelegt werden soll, existiert bereits!");
        }
        boolean success = createDirectory(dirName);
        if (!success) {
            throw new RuntimeException("Das Verzeichnis '" + dirName
                    + "', das angelegt werden soll, konnte nicht angelegt werden!");
        }
    }

    /**
     * Erstellt alle nicht vorhandenen Verzeichnisse zum gegebenen Dateinamen. Hinterher könnte man
     * dann die Datei kopieren, da das Verzeichnis existiert.
     *
     * @param filename
     *            Dateiname, zu dem ggf. fehlende Verzeichnisse angelegt werden sollen.
     */
    public static void createMissingDirectoriesForFile(String filename) {
        String path = getDirNameWithoutLastSlash(filename);
        createDirectoryWithMissingSubdirectoriesIfNotExists(path);
    }

    /**
     * Kopiert eine einzelne Datei mit NIO-Methoden.
     *
     * @param source
     *            Quelldatei
     * @param target
     *            Zieldatei
     * @throws IORuntimeException
     *             Falls etwas beim Kopieren schief geht.
     */
    public static void copyFile(String source, String target) {
        Path sourcePath = Paths.get(source);
        Path targetPath = Paths.get(target);
        try {
            Files.copy(sourcePath, targetPath, StandardCopyOption.REPLACE_EXISTING);
        }
        catch (IOException exception) {
            throw new IORuntimeException(exception);
        }
    }

    /**
     * Kopiert eine einzelne Datei mit NIO-Methoden.
     *
     * @param source
     *            Quelldatei
     * @param target
     *            Zieldatei
     * @throws IORuntimeException
     *             Falls etwas beim Kopieren schief geht.
     */
    public static void copyFile(File source, File target) {
        copyFile(source.getAbsolutePath(), target.getAbsolutePath());
    }

    /**
     * Kopiert eine einzelne Datei mit NIO-Methoden.
     *
     * @param source
     *            Quelldatei
     * @param target
     *            Zieldatei
     * @throws IORuntimeException
     *             Falls etwas beim Kopieren schief geht.
     */
    public static void copyFile(File source, String target) {
        copyFile(source.getAbsolutePath(), target);
    }

    /**
     * Verschiebt eine einzelne Datei mit NIO-Methoden.
     *
     * @param source
     *            Quelldatei
     * @param target
     *            Zieldatei
     * @throws IORuntimeException
     *             Falls etwas beim Verschieben schief geht.
     */
    public static void moveFile(String source, String target) {
        Path sourcePath = Paths.get(source);
        Path targetPath = Paths.get(target);
        try {
            Files.move(sourcePath, targetPath, StandardCopyOption.REPLACE_EXISTING);
        }
        catch (IOException exception) {
            throw new IORuntimeException(exception);
        }
    }

    /**
     * Verschiebt eine einzelne Datei mit NIO-Methoden.
     *
     * @param source
     *            Quelldatei
     * @param target
     *            Zieldatei
     * @throws IORuntimeException
     *             Falls etwas beim Verschieben schief geht.
     */
    public static void moveFile(File source, File target) {
        moveFile(source.getAbsolutePath(), target.getAbsolutePath());
    }

    /**
     * Verschiebt eine einzelne Datei (mit Hilfe von copyFile() und deleteFile()).
     *
     * @param source
     *            Quelldatei
     * @param target
     *            Zieldatei
     * @throws IORuntimeException
     *             Falls etwas beim Kopieren schief geht.
     */
    public static void moveFileViaCopyAndDelete(String source, String target) {
        copyFile(source, target);
        deleteFile(source);
    }

    /**
     * Verschiebt eine einzelne Datei (mit Hilfe von copyFile() und deleteFile()).
     *
     * @param source
     *            Quelldatei
     * @param target
     *            Zieldatei
     * @throws IORuntimeException
     *             Falls etwas beim Kopieren schief geht.
     */
    public static void moveFileViaCopyAndDelete(File source, File target) {
        moveFileViaCopyAndDelete(source.getAbsolutePath(), target.getAbsolutePath());
    }

    /**
     * Kopiert ein Verzeichnis mit Inhalt in ein anderes.
     *
     * @param source
     *            Quell-Verzeichnis.
     * @param target
     *            Ziel-Verzeichnis.
     * @throws IORuntimeException
     *             Wenn sich eine Datei oder Verzeichnis nicht kopieren oder anlegen lässt.
     */
    public static void copyTree(String source, String target) {
        copyTree(new File(source), new File(target));
    }

    /**
     * Kopiert ein Verzeichnis mit Inhalt in ein anderes.
     *
     * @param source
     *            Quell-Verzeichnis.
     * @param target
     *            Ziel-Verzeichnis.
     * @throws IORuntimeException
     *             Wenn sich eine Datei oder Verzeichnis nicht kopieren oder anlegen lässt.
     */
    public static void copyTree(File source, String target) {
        copyTree(source, new File(target));
    }

    /**
     * Kopiert ein Verzeichnis mit Inhalt in ein anderes.
     *
     * @param source
     *            Quell-Verzeichnis.
     * @param target
     *            Ziel-Verzeichnis.
     * @throws IORuntimeException
     *             Wenn sich eine Datei oder Verzeichnis nicht kopieren oder anlegen lässt.
     */
    public static void copyTree(File source, File target) {
        if (!source.exists()) {
            return;
        }
        for (File file : source.listFiles()) {
            if (file.isDirectory()) {
                String dirname = getBareName(file);
                String targetDir = concatPathes(target, dirname);
                createNotExistingDirectory(targetDir);
                copyTree(file, targetDir);
            }
            else {
                String dirname = getBareName(file);
                String targetFilename = concatPathes(target, dirname);
                copyFile(file, targetFilename);
            }
        }
    }

    /**
     * Prüft, ob die Datei mit dem gegebenen Dateinamen existiert.
     *
     * @param filename
     *            Name der Datei.
     * @return Wahrheitswert.
     */
    public static boolean exists(String filename) {
        File file = new File(filename);
        return file.exists();
    }

    /**
     * Prüft, ob die Datei mit dem gegebenen Dateinamen existiert und eine normale Date ist.
     *
     * @param filename
     *            Name der Datei.
     * @return Wahrheitswert.
     */
    public static boolean isFile(String filename) {
        File file = new File(filename);
        return file.isFile();
    }

    /**
     * Prüft, ob das Verzeichnis mit dem gegebenen Dateinamen existiert und ein Verzeichnis ist.
     *
     * @param dir
     *            Name des Verzeichnisses.
     * @return Wahrheitswert.
     */
    public static boolean isDirectory(String dir) {
        File file = new File(dir);
        return file.isDirectory();
    }

    /**
     * Prüft, ob das Verzeichnis mit dem gegebenen Dateinamen nicht existiert oder kein Verzeichnis
     * ist.
     *
     * @param dir
     *            Name des Verzeichnisses.
     * @return Wahrheitswert.
     */
    public static boolean isNotADirectory(String dir) {
        return !isDirectory(dir);
    }

    /**
     * Löscht eine Datei.
     *
     * @param filename
     *            Zu löschende Datei.
     * @throws IORuntimeException
     *             Falls sich die Datei nicht löschen lässt.
     */
    public static void deleteFile(String filename) {
        File file = new File(filename);
        if (!file.delete()) {
            throw new IORuntimeException(file + " could not be deleted!");
        }
    }

    /**
     * Löscht eine Datei, falls sie existiert.
     *
     * @param filename
     *            Zu löschende Datei.
     * @throws IORuntimeException
     *             Falls sich die Datei nicht löschen lässt.
     */
    public static void deleteFileIfExistent(String filename) {
        if (exists(filename)) {
            deleteFile(filename);
        }
    }

    /**
     * Legt eine Datei leer an. Wenn es sie schon gibt, wird sie leer überschrieben.
     *
     * @param filename
     *            Name der Datei.
     */
    public static void createEmptyFile(String filename) {
        FineFileWriter writer = new FineFileWriter(filename);
        writer.close();
    }

    /**
     * Erzeugt eine Datei leer, falls sie nicht bereits existiert.
     *
     * @param filename
     *            Name der Datei.
     */
    public static void createEmptyFileIfNotExisting(String filename) {
        if (!exists(filename)) {
            createEmptyFile(filename);
        }
    }

    /**
     * Legt eine Datei leer an. Wenn es sie schon gibt, wird sie leer überschrieben.
     *
     * @param filename
     *            Name der Datei.
     * @param charset
     *            Zeichensatz der zu schreibenden Datei.
     */
    public static void createEmptyFile(String filename, Charset charset) {
        FineFileWriter writer = new FineFileWriter(filename, charset);
        writer.close();
    }

    /**
     * Liest den Inhalt der Datei mit dem gegebenen Dateinamen in einen String und gibt diese
     * zurück.
     *
     * @param filename
     *            Name der Datei.
     * @return Inhalt der Datei.
     */
    public static String readFileToString(String filename) {
        Reader reader = new FineFileReader(filename);
        String contents = reader.readFileToString();
        reader.close();
        return contents;
    }

    /**
     * Liest den Inhalt der Datei im angegebenen Encoding mit dem gegebenen Dateinamen in einen
     * String und gibt diese zurück.
     *
     * @param filename
     *            Name der Datei.
     * @param charset
     *            Encoding der Datei.
     * @return Inhalt der Datei.
     */
    public static String readFileToString(String filename, Charset charset) {
        Reader reader = new FineFileReader(filename, charset);
        String contents = reader.readFileToString();
        reader.close();
        return contents;
    }

    /**
     * Liest den Inhalt der Datei mit dem gegebenen Dateinamen in einen String und gibt diese
     * zurück.
     *
     * @param file
     *            Datei.
     * @return Inhalt der Datei.
     */
    public static String readFileToString(File file) {
        return readFileToString(file.getPath());
    }

    /**
     * Liest den Inhalt der Datei mit dem gegebenen Dateinamen in eine Liste seiner Zeilen und gibt
     * diese zurück.
     *
     * @param filename
     *            Name der Datei.
     * @return Inhalt der Datei als Liste der Zeilen.
     */
    public static List<String> readFileToList(String filename) {
        return readFromReaderToList(new FineFileReader(filename));
    }

    /**
     * Liest die Zeilen aus dem übergebenen Reader in eine Liste ein und schließt den Reader dann.
     *
     * @param reader
     *            Zu verwendender Reader.
     * @return Liste mit den Zeilen.
     */
    private static List<String> readFromReaderToList(Reader reader) {
        List<String> contents = reader.readFileToListOfStrings();
        reader.close();
        return contents;
    }

    /**
     * Liest den Inhalt der Datei mit dem gegebenen Dateinamen in eine Liste
     * seiner Zeilen, wobei leere Zeilen und solche, die mit '#' anfangen,
     * übersprungen werden.                                                     <br><br>
     *
     * Die Zeilen werden nicht um vorn oder hinten befindliche Leerzeichen
     * bereinigt.
     *
     * @param filename
     *            Name der Datei.
     * @return Inhalt der Datei ohne Leerzeilen und Kommentare als Liste der Zeilen.
     */
    public static List<String> readFileToListIgnoreBlankAndCommentLines(String filename) {
        List<String> list = new ArrayList<>();

        Reader reader = new FineFileReader(filename);
        String line;
        while (null != (line = reader.readNextLine())) {
            String testLine = line.trim();
            if (!testLine.isEmpty() && !testLine.startsWith("#")) {
                list.add(line);
            }
        }

        return list;
    }

    /**
     * Liest den Inhalt der Datei mit dem gegebenen Dateinamen in eine Liste seiner Zeilen und gibt
     * diese zurück.
     *
     * @param filename
     *            Name der Datei.
     * @param charset
     *            Encoding der Datei.
     * @return Inhalt der Datei als Liste der Zeilen.
     */
    public static List<String> readFileToList(String filename, Charset charset) {
        return readFromReaderToList(new FineFileReader(filename, charset));
    }

    /**
     * Liest die letzte nicht leere Zeile aus der Datei.
     *
     * @param filename
     *            Name der Datei.
     * @param charset
     *            Encoding der Datei.
     * @return Inhalt der letzten nicht leeren Zeile aus der Datei oder der leere String.
     */
    public static String readLastNotEmptyLine(String filename) {
        List<String> list = readFileToList(filename);
        for (int index = list.size() - 1; index >= 0; --index) {
            String line = list.get(index);
            if (!line.isEmpty()) {
                return line;
            }
        }

        return "";
    }

    /**
     * Speichert eine Datei als Sicherheitskopie ab. Der Name wird dabei automatisch erzeugt.
     *
     * @param originalFileName
     *            Name der Originaldatei, die gesichert werden soll.
     */
    public static void backupFile(String originalFileName) {
        File originalFile = new File(originalFileName);
        if (originalFile.exists()) {
            long lastModified = originalFile.lastModified();
            GregorianCalendar cal = new GregorianCalendar();
            cal.setTime(new Date(lastModified));
            String insertion = String.format("_%1$tY%1$tm%1$td", cal);
            String backupFileName = insertBeforeExtension(originalFileName, insertion);

            copyFile(originalFileName, backupFileName);
        }
    }

    /**
     * Fügt einen String in einen Dateinamen vor den Punkt und die Erweiterung. Also wird aus
     * "/user/bla/datei.txt" und "_foo" dann ein "/user/bla/datei_foo.txt".
     *
     * @param originalFileName
     *            Der Ausgangsname.
     * @param insertion
     *            Der einzufügende Teil.
     * @return Name mit eingefügtem Text.
     */
    public static String insertBeforeExtension(String originalFileName, String insertion) {
        String dir = FileHelper.getDirName(originalFileName);
        String bareName = FileHelper.getBareName(originalFileName);
        int lastPointIndex = bareName.lastIndexOf(".");
        String newBareName;
        if (lastPointIndex > -1) {
            String beforePoint = bareName.substring(0, lastPointIndex);
            String pointAndAfter = bareName.substring(lastPointIndex);
            newBareName = beforePoint + insertion + pointAndAfter;
        }
        else {
            newBareName = bareName + insertion;
        }
        if (dir.isEmpty()) {
            return newBareName;
        }
        else {
            return concatPathes(dir, newBareName);
        }
    }

    /**
     * Entfernt die Extension ("bla.txt" -> "bla", "c:/foo/bar.txt" -> "c:/foo/bar").
     *
     * @param originalFileName
     *            Der Ausgangsname.
     * @return Name ohne Extension.
     */
    public static String removeExtension(String originalFileName) {
        int lastPointIndex = originalFileName.lastIndexOf(".");
        if (lastPointIndex > -1) {
            return originalFileName.substring(0, lastPointIndex);
        }
        else {
            return originalFileName;
        }
    }

    /**
     * Ermittelt die Extension ("bla.txt" -> ".txt", "c:/foo/bar.foo" -> ".foo").
     *
     * @param originalFileName
     *            Der Ausgangsname.
     * @return Extension mit führendem Punkt.
     */
    public static String determineExtension(String originalFileName) {
        int lastPointIndex = originalFileName.lastIndexOf(".");
        if (lastPointIndex > -1) {
            return originalFileName.substring(lastPointIndex);
        }
        else {
            return NO_EXTENSION_FOUND;
        }
    }

    /**
     * Speichert einen String in eine Datei.
     *
     * @param text
     *            Der zu speichernde String.
     * @param filename
     *            Der Name der zu erzeugenden Datei.
     */
    public static void writeTextToFile(String text, String filename) {
        FineFileWriter writer = new FineFileWriter(filename);
        writer.writeln(text);
        writer.close();
    }

    /**
     * Speichert einen String in eine Datei.
     *
     * @param text
     *            Der zu speichernde String.
     * @param filename
     *            Der Name der zu erzeugenden Datei.
     * @param charset
     *            Zeichensatz der zu schreibenden Datei.
     */
    public static void writeTextToFile(String text, String filename, Charset charset) {
        FineFileWriter writer = new FineFileWriter(filename, charset);
        writer.writeln(text);
        writer.close();
    }

    /**
     * Speichert einen String in eine Datei mit aktuellem Datum und Zeit als Einschub vor die
     * Erweiterung.
     *
     * @param text
     *            Der zu speichernde String.
     * @param front
     *            Der vordere Teil des Namens der zu erzeugenden Datei.
     * @param rear
     *            Der hintere Teil des Namens der zu erzeugenden Datei.
     */
    public static void writeTextToFileWithDate(String text, String front, String rear) {
        GregorianCalendar cal = new GregorianCalendar();
        String insertion = String.format("%1$tY%1$tm%1$td_%1$tH%1$tM%1$tS", cal);
        writeTextToFile(text, front + insertion + rear);
    }

    /**
     * Schreibt die Liste mit den Zeilen in die Datei mit dem angegebenen Namen.
     *
     * @param lines
     *            Die Liste mit den auszugebenden Zeilen.
     * @param filename
     *            Name der zu erzeugenden Datei.
     */
    public static void writeLinesToFile(List<String> lines, String filename) {
        writeLinesToWriter(new FineFileWriter(filename), lines);
    }

    /**
     * Schreibt die Liste mit den Zeilen unter Verwendung des angegebenen Encodings in die Datei
     * mit dem angegebenen Namen.
     *
     * @param lines
     *            Die Liste mit den auszugebenden Zeilen.
     * @param filename
     *            Name der zu erzeugenden Datei.
     * @param charset
     *            Encoding der Datei.
     */
    public static void writeLinesToFile(List<String> lines, String filename, Charset charset) {
        writeLinesToWriter(new FineFileWriter(filename, charset), lines);
    }

    /**
     * Schreibt den Text im Format "UCS-2 LE BOM".
     * Dann erkennt Notepad++ das auch als "UCS-2 LE BOM".
     *
     * @param lines
     *            Die Liste mit den auszugebenden Zeilen.
     * @param filename
     *            Name der zu erzeugenden Datei.
     * @see http://stackoverflow.com/questions/17081953/write-text-file-in-ucs-2-little-endian-java
     */
    public static void writeLinesToUCS2LeBom(List<String> lines, String filename) {
        writeLinesToFileWithByteOrderMask(lines, filename, Charset.UTF_16LE, "\uFEFF");
    }

    /**
     * Schreibt den Text im Format "UTF-8-BOM".
     * Dann erkennt Notepad++ das auch als "UTF-8-BOM".
     *
     * @param lines
     *            Die Liste mit den auszugebenden Zeilen.
     * @param filename
     *            Name der zu erzeugenden Datei.
     * @see http://stackoverflow.com/questions/4389005/how-to-add-a-utf-8-bom-in-java
     */
    public static void writeLinesToUTF8Bom(List<String> lines, String filename) {
        writeLinesToFileWithByteOrderMask(lines, filename, Charset.UTF_8, "\uFEFF");
    }

    /**
     * Schreibt die Liste mit den Zeilen unter Verwendung des angegebenen Encodings und der
     * angegebenen Byte-Order-Mask in die Datei mit dem angegebenen Namen.
     *
     * Dummer Weise ist die Datei nicht im richtigen Format, wenn man sie mit dem Charset.UTF_16LE,
     * also "UTF-16LE" bzw. "Sixteen-bit UCS Transformation Format, little-endian byte order")
     * schreibt.
     *
     * Laut http://stackoverflow.com/questions/17081953/write-text-file-in-ucs-2-little-endian-java
     * muss man am Anfang
     *     writer.write("\uFEFF");
     * ausführen. Das wird hier getan.
     *
     * @param lines
     *            Die Liste mit den auszugebenden Zeilen.
     * @param filename
     *            Name der zu erzeugenden Datei.
     * @param charset
     *            Encoding der Datei.
     * @param byteOrderMask
     *            Zu schreibende Byte-Order-Mask.
     */
    public static void writeLinesToFileWithByteOrderMask(List<String> lines, String filename,
            Charset charset, String byteOrderMask) {
        Writer writer = new FineFileWriter(filename, charset);
        writer.write(byteOrderMask);
        writeLinesToWriter(writer, lines);
    }

    /**
     * Schreibt die Zeilen auf den angegebenen Writer und schließt diesen dann.
     *
     * @param writer
     *            Zu verwendender Writer.
     * @param lines
     *            Auszugebende Zeilen.
     */
    private static void writeLinesToWriter(Writer writer, List<String> lines) {
        writer.writeAllLines(lines);
        writer.close();
    }

    /**
     * Schreibt den Text im Format "UCS-2 LE BOM".
     * Dann erkennt Notepad++ das auch als "UCS-2 LE BOM".
     *
     * @param text
     *            Auszugebender Text.
     * @param filename
     *            Name der zu erzeugenden Datei.
     * @see http://stackoverflow.com/questions/17081953/write-text-file-in-ucs-2-little-endian-java
     */
    public static void writeTextToUCS2LeBom(String text, String filename) {
        writeTextToFileWithByteOrderMask(text, filename, Charset.UTF_16LE, "\uFEFF");
    }

    /**
     * Schreibt den Text im Format "UTF-8-BOM".
     * Dann erkennt Notepad++ das auch als "UTF-8-BOM".
     *
     * @param text
     *            Auszugebender Text.
     * @param filename
     *            Name der zu erzeugenden Datei.
     * @see http://stackoverflow.com/questions/4389005/how-to-add-a-utf-8-bom-in-java
     */
    public static void writeTextToUTF8Bom(String text, String filename) {
        writeTextToFileWithByteOrderMask(text, filename, Charset.UTF_8, "\uFEFF");
    }

    /**
     * Schreibt den Text unter Verwendung des angegebenen Encodings und der angegebenen
     * Byte-Order-Mask in die Datei mit dem angegebenen Namen.
     *
     * Dummer Weise ist die Datei nicht im richtigen Format, wenn man sie mit dem Charset.UTF_16LE,
     * also "UTF-16LE" bzw. "Sixteen-bit UCS Transformation Format, little-endian byte order")
     * schreibt.
     *
     * Laut http://stackoverflow.com/questions/17081953/write-text-file-in-ucs-2-little-endian-java
     * muss man am Anfang
     *     writer.write("\uFEFF");
     * ausführen. Das wird hier getan.
     *
     * @param text
     *            Der auszugebende Text.
     * @param filename
     *            Name der zu erzeugenden Datei.
     * @param charset
     *            Encoding der Datei.
     * @param byteOrderMask
     *            Zu schreibende Byte-Order-Mask.
     */
    public static void writeTextToFileWithByteOrderMask(String text, String filename,
            Charset charset, String byteOrderMask) {
        Writer writer = new FineFileWriter(filename, charset);
        writer.write(byteOrderMask);
        writer.writeln(text);
        writer.close();
    }

    /**
     * Fügt den angegebenen Text und einen Zeilenumbruch an die angegebene Datei an.
     *
     * @param line
     *            Anzufügender Text.
     * @param filename
     *            Dateiname
     */
    public static void appendLineToFile(String line, String filename) {
        Writer writer = new FineFileWriter(filename, true);
        writer.writeln(line);
        writer.close();
    }

    /**
     * Fügt den angegebenen Text und einen Zeilenumbruch an die angegebene Datei an.
     *
     * @param line
     *            Anzufügender Text.
     * @param filename
     *            Dateiname
     * @param charset
     *            Zu verwendendes Charset.
     */
    public static void appendLineToFile(String line, String filename, Charset charset) {
        Writer writer = new FineFileWriter(filename, charset, true);
        writer.writeln(line);
        writer.close();
    }

    /**
     * Verschiebt eine Datei (mit Hilfe von {@link File#renameTo(File)}).
     *
     * @param source
     *            Die Datei, die kopiert werden soll.
     * @param target
     *            Die Datei, die erstellt und geschrieben wird.
     * @return Erfolg
     */
    public static boolean moveFileByRenameTo(String source, String target) {
        if (source.equals(target)) {
            return false;
        }

        File sourceFile = new File(source);
        File targetFile = new File(target);
        return sourceFile.renameTo(targetFile);
    }

    /**
     * Kopiert den Text aus einer Datei in eine andere.
     *
     * @param source
     *            Die Datei, die kopiert werden soll.
     * @param target
     *            Die Datei, die erstellt und geschrieben wird.
     */
    public static void copyTextFileByReadAndWrite(String source, String target) {
        if (source.equals(target)) {
            return;
        }

        Reader reader = new FineFileReader(source);
        Writer writer = new FineFileWriter(target);
        while (true) {
            String line = reader.readNextLine();
            if (null == line) {
                break;
            }
            writer.writeln(line);
        }
        writer.close();
        reader.close();
    }

    /**
     * Setzt zwei Pfade so zusammen, dass sowohl doppelte als auch fehlende Trenner vermieden
     * werden.
     *
     * @param part1
     *            Erster Teil des Pfades.
     * @param part2
     *            Zweiter Teil des Pfades.
     * @return Zusammengesetzter Pfad.
     */
    public static String concatPathes(String part1, String part2) {
        boolean endsWithSlash = part1.endsWith("/") || part1.endsWith("\\");
        boolean startsWithSlash = part2.startsWith("/") || part2.startsWith("\\");

        if (endsWithSlash && startsWithSlash) {
            String tmp = part1.substring(0, part1.length() - 1);
            return tmp + part2;
        }
        else if (endsWithSlash || startsWithSlash) {
            return part1 + part2;
        }
        else {
            return part1 + File.separator + part2;
        }
    }

    /**
     * Setzt zwei Pfade so zusammen, dass sowohl doppelte als auch fehlende Trenner vermieden
     * werden.
     *
     * @param part1
     *            Erster Teil des Pfades.
     * @param part2
     *            Zweiter Teil des Pfades.
     * @return Zusammengesetzter Pfad.
     */
    public static String concatPathes(File part1, String part2) {
        return concatPathes(part1.getPath(), part2);
    }

    /**
     * Setzt zwei Pfade so zusammen, dass sowohl doppelte als auch fehlende Trenner vermieden
     * werden.
     *
     * @param part1
     *            Erster Teil des Pfades.
     * @param part2
     *            Zweiter Teil des Pfades.
     * @return Zusammengesetzter Pfad.
     */
    public static String concatPathes(String part1, File part2) {
        return concatPathes(part1, part2.getPath());
        /*
         * Macht das Sinn? Dann würde man an irgendwas einen kompletten Pfad dranhängen, vermutlich
         * nicht! Momentan wird die Methode auch nicht genutzt.
         */
    }

    /**
     * Setzt zwei Pfade so zusammen, dass sowohl doppelte als auch fehlende Trenner vermieden
     * werden.
     *
     * @param part1
     *            Erster Teil des Pfades.
     * @param part2
     *            Zweiter Teil des Pfades.
     * @return Zusammengesetzter Pfad.
     */
    public static String concatPathes(File part1, File part2) {
        return concatPathes(part1.getPath(), part2.getPath());
        /*
         * Macht das Sinn? Dann würde man an einen kompletten Pfad einen kompletten Pfad
         * dranhängen, vermutlich nicht! Momentan wird die Methode auch nicht genutzt.
         */
    }

    /**
     * Setzt mehrere Pfadstücke so zusammen, dass sowohl doppelte als auch fehlende Trenner
     * vermieden werden.
     *
     * @param parts
     *            Teile des Pfades
     * @return Zusammengesetzter Pfad.
     */
    public static String concatPathesMultiple(String ... parts) {
        if (parts.length == 0) {
            return "";
        }
        else if (parts.length == 1) {
            return parts[0];
        }

        String filename = parts[0];
        for (int index = 1; index < parts.length; ++index) {
            filename = concatPathes(filename, parts[index]);
        }

        return filename;
    }

    /**
     * Erzeugt aus einer Zeile der Art                          <br><br>&nbsp;&nbsp;&nbsp;&nbsp;<tt>
     *     "DokNr";"Vortext"                                    </tt><br><br>
     * eine Liste mit den einzelnen Werten.
     *
     * @param line
     *            Eingelesene Zeile.
     * @return Liste mit den Werten der Zeile.
     */
    public static List<String> splitCsvSemikolonLine(String line) {
        if (!line.startsWith("\"")) {
            throw new IllegalArgumentException("Die zum Aufsplitten übergebene Zeile beginnt "
                    + "nicht mit einem Anführungzeichen!\n\tZeile: " + line);
        }
        if (!line.endsWith("\"")) {
            throw new IllegalArgumentException("Die zum Aufsplitten übergebene Zeile wird nicht "
                    + "von einem Anführungzeichen beendet!\n\tZeile: " + line);
        }

        /* Vorderes und hinteres Anführungszeichen abschneiden: */
        String shortLine = line.substring(1, line.length() - 1);

        return Arrays.asList(shortLine.split("\";\"", -1));
    }

    /**
     * Ermittelt alle Dateien in einem Verzeichnis, die einem bestimmten Kriterium entsprechen.
     *
     * @param source
     *            Verzeichnis, in dem gesucht wird.
     * @param filenameFilter
     *            Kriterium an den Dateinamen der zu suchenden Dateien.
     * @return Liste mit allen passenden Dateien.
     */
    public static List<String> searchDir(File source, FilenameFilter filenameFilter) {
        List<String> files = new ArrayList<>();

        for (File file : source.listFiles(filenameFilter)) {
            files.add(file.getPath());
        }

        return files;
    }

    /**
     * Ermittelt den letzten Zeitpunkt der Modifikation einer Datei.
     *
     * @param filename
     *            Name der Datei.
     * @return Zeitpunkt im Timestamp-Format YYYYMMDD_HHMMSS.
     */
    public static String fileLastModified(String filename) {
        return fileLastModified(new File(filename));
    }

    /**
     * Ermittelt den letzten Zeitpunkt der Modifikation einer Datei.
     *
     * @param file
     *            Zu untersuchende Datei.
     * @return Zeitpunkt im Timestamp-Format YYYYMMDD_HHMMSS.
     */
    public static String fileLastModified(File file) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd_HHmmss");
        return sdf.format(fileLastModifiedAsMillis(file));
    }

    /**
     * Ermittelt den letzten Zeitpunkt der Modifikation einer Datei.
     *
     * @param filename
     *            Name der Datei.
     * @return Zeitpunkt als Millisekunden.
     */
    public static long fileLastModifiedAsMillis(String filename) {
        return fileLastModifiedAsMillis(new File(filename));
    }

    /**
     * Ermittelt den letzten Zeitpunkt der Modifikation einer Datei.
     *
     * @param file
     *            Zu untersuchende Datei.
     * @return Zeitpunkt als Millisekunden.
     */
    public static long fileLastModifiedAsMillis(File file) {
        return file.lastModified();
    }

    /** Setzt den Zeitpunkt der letzten Modifikation der angegebenen Datei. */
    public static void setLastModified(String filename, long lastModified) {
        setLastModified(new File(filename), lastModified);
    }

    /** Setzt den Zeitpunkt der letzten Modifikation der angegebenen Datei. */
    private static void setLastModified(File file, long lastModified) {
        if (file.isFile()) {
            file.setLastModified(lastModified);
        }
    }

    /**
     * Stellt fest, ob das angegebene Verzeichnis leer ist.
     *
     * @param dir
     *            Zu überprüfendes Verzeichnis, muss existieren.
     * @return Wahrheitswert: true genau dann, wenn das Verzeichnis leer ist.
     */
    public static boolean isDirectoryEmpty(String dir) {
        File dirAsFile = new File(dir);
        int numberOfFoundFiles = dirAsFile.listFiles().length;
        return 0 == numberOfFoundFiles;
    }

    /**
     * Stellt fest, ob das angegebene Verzeichnis ein Verzeichnis und leer ist.
     *
     * @param dir
     *            Zu überprüfendes Verzeichnis, muss existieren.
     * @return Wahrheitswert: true genau dann, wenn das Verzeichnis ein Verzeichnis und leer ist.
     */
    public static boolean isEmptyDirectory(String dir) {
        return isDirectory(dir) && isDirectoryEmpty(dir);
    }

    /**
     * Stellt fest, ob das angegebenen Verzeichnis Unterverzeichnisse enthält.
     *
     * @param dir
     *            Zu überprüfendes Verzeichnis, muss existieren.
     * @return Wahrheitswert: true genau dann, wenn das Verzeichnis Unterverzeichnisse enthält.
     */
    public static boolean hasDirectorySubdirectories(String dir) {
        if (!isDirectory(dir)) {
            return false;
        }
        File dirAsFile = new File(dir);
        for (File file : dirAsFile.listFiles()) {
            if (file.isDirectory()) {
                return true;
            }
        }
        return false;
    }

    /**
     * Erstellt ein leere Datei im Default-Verzeichnis für temporäre Dateien, die mit dem Prefix
     * beginnt und mit dem Suffix endet.
     *
     * @param prefix
     *            Prefix mit dem der Dateiname beginnt. Muss mindestens drei Zeichen lang sein.
     * @param suffix
     *            Suffix auf den der Dateiname endet.
     * @return File-Objekt mit der erstellten Datei.
     * @throws IORuntimeException
     *             Falls etwas dabei schief geht.
     */
    public static File createTempFile(String prefix, String suffix) {
        try {
            return File.createTempFile(prefix, suffix);
        }
        catch (IOException e) {
            throw new IORuntimeException(e);
        }
    }

    /**
     * Entfernt . und .. aus Dateinamen.
     *
     * @param filename
     *            Dateiname.
     * @return normalisierter Dateiname.
     */
    public static String normalisePath(String filename) {
        try {
            File file = new File(filename);
            return file.getCanonicalPath();
        }
        catch (IOException e) {
            throw new IORuntimeException(e);
        }
    }

    /**
     * Ermittelt die Laufwerksbezeichnung zu einer beliebigen Datei oder einem Verzeichnis.
     *
     * @param path
     *            Dateiname oder Verzeichnisname mit vollem Pfad.
     * @return Laufwerksbezeichnung
     * @throws RuntimeException
     *             Falls path nicht existiert oder zu path die Laufwerksbezeichnung nicht ermittelt
     *             werden konnte (z.B. wenn path nicht mit "R:\\" oder "r:\\" oder so beginnt,
     *             sondern mit "R:/" (hier hilft normalizeToBackSlashes()) oder
     *             "\\Cluster1\Projekte1").
     */
    public static String getSystemDisplayName(String path) {
        File file = new File(path);
        if (!file.exists()) {
            throw new RuntimeException("Die Datei oder das Verzeichnis '" + path
                    + "' existiert nicht!");
        }

        FileSystemView fsv = FileSystemView.getFileSystemView();
        String upperCasePath = path.toUpperCase();
        for (File rootAsFile : File.listRoots()) {
            String root = rootAsFile.getPath();
            String upperCaseRoot = root.toUpperCase();
            if (upperCasePath.startsWith(upperCaseRoot)) {
                String displayName = fsv.getSystemDisplayName(rootAsFile);
                if (displayName.matches(".* \\([A-Z]:\\)")) {
                    displayName = displayName.substring(0, displayName.length() - 5);
                }
                return displayName;
            }
        }

        throw new RuntimeException("Zur Datei oder das Verzeichnis '" + path
                    + "' konnte die Laufwerksbezeichnung nicht ermittelt werden!");
    }

    /**
     * Macht aus einem Pfad der Art "C:\Temp" einen String, der als Name einer Datei gültig ist.
     * Im Beispiel "c_Temp".
     *
     * @param path Pfadangabe mit Verzeichnistrennern, Laufwerksbuchstaben etc.
     * @return Name der als Dateiname (ohne Pfad) gültig wäre.
     */
    public static String buildFilenameFromPath(String path) {
        String file = path;

        file = file.replace(":", "");
        file = file.replace("\\", "_");
        file = file.replace("/", "_");
        file = file.replace(";", "_");
        file = file.replaceAll("^_+", "");

        return file;
    }

    /**
     * Ermittelt die Länge der Datei in Bytes.
     *
     * @param file
     *            Datei.
     * @return Länge der Datei in Bytes.
     * @throws IORuntimeException
     *             Falls die Datei nicht existiert oder ein Verzeichnis ist.
     */
    public static long fileSize(File file) {
        if (!file.isFile()) {
            throw new IORuntimeException("Die Datei '" + file.getPath()
                    + "', deren Länge ermittelt wird, ist keine Datei im Dataeisystem!");
        }
        return file.length();
    }

    /**
     * Ermittelt die Länge der Datei in Bytes.
     *
     * @param filename
     *            Name der Datei mit Pfad.
     * @return Länge der Datei in Bytes.
     * @throws IORuntimeException
     *             Falls die Datei nicht existiert oder ein Verzeichnis ist.
     */
    public static long fileSize(String filename) {
        return fileSize(new File(filename));
    }

    /**
     * Prüft ob zwei Dateien die gleiche Last-Modified-Time und die gleiche Dateigröße haben.
     *
     * @param fileA
     *            Datei 1.
     * @param fileB
     *            Datei 2.
     * @return Wahrheitswert.
     */
    public static boolean filesAreEqualEnough(File fileA, File fileB) {
        return filesHaveSameLastModifiedTime(fileA, fileB)
                && filesHaveEqualSize(fileA, fileB);
    }

    /**
     * Prüft ob zwei Dateien die gleiche Last-Modified-Time und die gleiche Dateigröße haben.
     *
     * @param filenameA
     *            Name der ersten Datei.
     * @param filenameB
     *            Name der zweiten Datei.
     * @return Wahrheitswert.
     */
    public static boolean filesAreEqualEnough(String filenameA, String filenameB) {
        return filesAreEqualEnough(new File(filenameA), new File(filenameB));
    }

    /**
     * Prüft ob zwei Dateien die gleiche Last-Modified-Time haben.
     *
     * @param fileA
     *            Datei 1.
     * @param fileB
     *            Datei 2.
     * @return Wahrheitswert.
     */
    public static boolean filesHaveSameLastModifiedTime(File fileA, File fileB) {
        String lastModifiedA = fileLastModified(fileA);
        String lastModifiedB = fileLastModified(fileB);
        return lastModifiedA.equals(lastModifiedB);
    }

    /**
     * Prüft ob zwei Dateien die gleiche Last-Modified-Time haben.
     *
     * @param filenameA
     *            Name der ersten Datei.
     * @param filenameB
     *            Name der zweiten Datei.
     * @return Wahrheitswert.
     */
    public static boolean filesHaveSameLastModifiedTime(String filenameA, String filenameB) {
        return filesHaveSameLastModifiedTime(new File(filenameA), new File(filenameB));
    }

    /**
     * Prüft ob zwei Dateien die gleiche gleiche Dateigröße haben.
     *
     * @param fileA
     *            Datei 1.
     * @param fileB
     *            Datei 2.
     * @return Wahrheitswert.
     */
    public static boolean filesHaveEqualSize(File fileA, File fileB) {
        long sizeA = fileSize(fileA);
        long sizeB = fileSize(fileB);
        return sizeA == sizeB;
    }

    /**
     * Prüft ob zwei Dateien die gleiche gleiche Dateigröße haben.
     *
     * @param filenameA
     *            Name der ersten Datei.
     * @param filenameB
     *            Name der zweiten Datei.
     * @return Wahrheitswert.
     */
    public static boolean filesHaveEqualSize(String filenameA, String filenameB) {
        return filesHaveEqualSize(new File(filenameA), new File(filenameB));
    }

    /** Tauscht die Endung eines Dateinamens aus. */
    public static String exchangeExtension(String filename, String newExtension) {
        int lastPointIndex = filename.lastIndexOf(".");
        if (0 > lastPointIndex) {
            throw new IllegalArgumentException("Der Dateiname '" + filename
                    + "' enthält keinen Punkt.");
        }
        if (0 == lastPointIndex) {
            throw new IllegalArgumentException("Der Dateiname '" + filename
                    + "' startet mit einem Punkt und das ist auch der einzige Punkt.");
        }

        String possibleOldExtension = filename.substring(lastPointIndex + 1);
        if (possibleOldExtension.isEmpty()) {
            throw new IllegalArgumentException("Der Dateiname '" + filename
                    + "' endet auf einen Punkt.");
        }

        if (Text.containsOnlyNormalLetters(possibleOldExtension)) {
            String partWithoutExtension = filename.substring(0, lastPointIndex);
            String newFilename;
            if (newExtension.startsWith(".")) {
                newFilename = partWithoutExtension + newExtension;
            }
            else {
                newFilename = partWithoutExtension + "." + newExtension;
            }
            if (filename.equals(newFilename)) {
                throw new IllegalArgumentException("Der neue Dateiname ist gleich dem übergebenen "
                        + "Dateinamen!\n\t"
                        + "filename     = " + filename + "\n\t"
                        + "newExtension = " + newExtension + "\n\t"
                        + "newFilename  = " + newFilename);
            }
            else {
                return newFilename;
            }
        }
        else {
            throw new IllegalArgumentException("Die Endung '" + possibleOldExtension
                    + "' des Dateinamens '" + filename + "' ist seltsam.");
        }
    }

    /**
     * Ersetzt einige UNC-Pfade durch die Laufwerksbuchstaben, da sonst Lynx damit nicht klarkommt.
     *
     * @param path
     *            Zu bearbeitender Pfad.
     * @return Bearbeiteter Pfad.
     */
    public static String replaceUncPathes(String path) {
        String replacedPath = path;

        replacedPath = replacedPath.replace("\\\\Cluster1\\user1\\CD", "u:");
        replacedPath = replacedPath.replace("\\\\Cluster1\\User1\\CD", "u:");

        replacedPath = replacedPath.replace("\\\\Cluster2\\Projekte2\\Schufa\\Impressum", "i:");
        replacedPath = replacedPath.replace("\\\\Cluster1\\Projekte1", "r:");
        replacedPath = replacedPath.replace("\\\\Cluster2\\Projekte2", "s:");

        return replacedPath;
    }

    /**
     * Liest eine Datei ein und entfernt die Einrückungen, die Lynx erzeugt.
     *
     * @param filename
     *            Name der einzulesenden Datei.
     * @param charset
     *            Das Charset der einzulesenden Datei.
     * @return String mit dem eingelesenen Inhalt der Datei.
     */
    public static String readAndRemoveIndentation(String filename, Charset charset) {
        Reader reader = new FineFileReader(filename, charset);
        StringBuilder builder = new StringBuilder();

        String line = reader.readNextLine();
        while (line != null) {
            builder.append(line.trim());
            builder.append(Text.LINE_BREAK);
            line = reader.readNextLine();
        }

        reader.close();
        return builder.toString();
    }

    /** Löscht alle Dateien und Unterverzeichnisse im angegebenen Verzeichnis. */
    public static void cleanDirectory(String directory) {
        deleteTree(directory);
        createDirectory(directory);
    }

    /**
     * Erstellt eine geordnete Liste aller Dateien in einem Verzeichnis (ohne Unterverzeichnisse zu
     * beachten), die Dateien sind dann absteigend nach dem letzten Zeitpunkt der Modifikation
     * einer Datei sortiert.
     *
     * @param dir
     *            Verzeichnis dessen Dateien sortiert zurückgegeben werden sollen.
     * @return Absteigend sortierte Liste der Dateien.
     */
    public static List<FileWithModificationTime> determineAllFilesInMainDirectoryDescendingByTime(
            String dir) {
        List<String> files = findAllFilesInMainDirectoryNio2(dir);
        List<FileWithModificationTime> fileWithModificationTimes = new ArrayList<>();

        for (String file : files) {
            FileWithModificationTime fileWithModificationTime = new FileWithModificationTime(file);
            fileWithModificationTimes.add(fileWithModificationTime);
        }

        Collections.sort(fileWithModificationTimes, new Comparator<FileWithModificationTime>() {
            @Override
            public int compare(FileWithModificationTime file1, FileWithModificationTime file2) {
                String modified1 = file1.getModificationTime();
                String modified2 = file2.getModificationTime();

                int cmp = modified2.compareTo(modified1);
                if (cmp == 0) {
                    String filename1 = file1.getFilename();
                    String filename2 = file2.getFilename();
                    cmp = filename1.compareTo(filename2);
                }
                return cmp;
            }
        });

        return fileWithModificationTimes;
    }

    public static String prepareToFilenamePart(String input) {
        String filenamePart = input;

        filenamePart = filenamePart.replaceAll("[^_.A-Za-z0-9ÄÖÜäöüß-]", "_");
        filenamePart = filenamePart.replaceAll("__+", "_");
        filenamePart = filenamePart.replaceAll("^_+", "");
        filenamePart = filenamePart.replaceAll("_+$", "");

        return filenamePart;
    }

    /** Legt eine Datei leer an oder überschreibt eine vorhandene Datei leer. */
    public static void clearFile(String filename, Charset charset) {
        Writer writer = new FineFileWriter(filename, charset);
        writer.close();
    }

}
