package de.duehl.basics.logic;

/*
 * Copyright 2016 Christian Dühl. All rights reserved.
 *
 * This program is free software. You can redistribute it and/or
 * modify it under the same terms as perl:
 *
 * general:  http://dev.perl.org/licenses/
 * GPL:      http://dev.perl.org/licenses/gpl1.html
 * artistic: http://dev.perl.org/licenses/artistic.html
 */

import java.util.ArrayList;
import java.util.List;

/**
 * Diese Klasse dient Informationsquelle über Datum und Version des Programms.
 *
 * @version 1.03     2017-02-01
 * @author Christian Dühl
 */

public final class Version implements Comparable<Version> {

    /** Version als String. */
    private final String version;

    /** Datum als String. */
    private final String date;

    /** Interne Darstellung der Version als Liste von Integern. */
    private final List<Integer> numericVersion;

    /**
     * Konstruktor
     *
     * @param version
     *            Version als String.
     * @param date
     *            Datum als String.
     */
    public Version(final String version, final String date) {
        this.date = date;
        this.version = version;
        numericVersion = new ArrayList<>();
        initNumericVersion(version);
    }

    /**
     * Erzeugt aus der String-Version eine Liste ihrer interessanten Teile in
     * Integer-Representation.
     *
     * @param versionString
     *            Version als String
     */
    private void initNumericVersion(String versionString) {
        /* 1. Aufteilung an nicht-(Zahlen und Buchstaben): */
        String[] preParts = versionString.split("[^0-9a-zA-Z]+");

        /* 2. Teile in Zahlen und einzelne Buchstaben aufteilen: */
        for (String prePart : preParts) {
            /* 2.1 reine Ziffernfolge: */
            if (prePart.matches("[0-9]+")) {
                int number = Integer.parseInt(prePart);
                numericVersion.add(number);
            }
            /* 2.2 reine alphanumerische Zeichen: */
            else if (prePart.matches("[a-zA-Z]+")) {
                char[] chars = prePart.toCharArray();
                for (char c : chars) {
                    int i = c;
                    numericVersion.add(i);
                }
            }
            /* 2.3 gemischte Zeichen: */
            else {
                 boolean numerical = false;
                 char[] chars = prePart.toCharArray();
                 StringBuilder builder = new StringBuilder();
                 for (char c : chars) {
                     String str = String.valueOf(c);
                     if (str.matches("[a-zA-Z]")) {
                         if (numerical) {
                             numerical = false;
                             String numberString = builder.toString();
                             int number = Integer.parseInt(numberString);
                             numericVersion.add(number);
                             builder.delete(0, builder.length());
                         }
                         int i = c;
                         numericVersion.add(i);
                     }
                     else {
                         numerical = true;
                         builder.append(c);
                     }
                 }
                /*
                 * Falls am Ende noch eine ungesicherte Zahlenfolge da ist,
                 * diese speichern:
                 */
                 if (numerical) {
                     String numberString = builder.toString();
                     int number = Integer.parseInt(numberString);
                     numericVersion.add(number);
                 }
            }
        }
    }

    /** Getter für die Version als String. */
    public String getVersion() {
        return version;
    }

    /** Getter für den Test für die numerische Version. */
    protected List<Integer> getNumericVersion() {
        return numericVersion;
    }

    /** Getter für das Datum als String. */
    public String getDate() {
        return date;
    }

    /** Getter für die Version und das Datum als zusammengesetzter String. */
    public String getVersionAndDate() {
        return version + " vom " + date;
    }

    /** Stringrepräsentation. */
    @Override
    public String toString() {
        return "Version " + version + " vom " + date;
    }

    /**
     * Vergleich mit einem zweiten Versions-Objekt.                               <br><br>
     *
     * Wichtig ist neben einer Implementierung von compareTo() auch die passende
     * Realisierung in equals(). Sie ist erst dann konsistent, wenn
     * e1.compareTo(e2) == 0 das gleiche Ergebnis wie e1.equals(e2) liefert,
     * wobei e1 und e2 den gleichen Typ besitzen. Ein Verstoß gegen diese Regel
     * kann bei sortierten Mengen schnell Probleme bereiten; ein Beispiel nennt
     * die API-Dokumentation. Auch sollte die hashCode()-Methode korrekt
     * realisiert sein, denn sind Objekte gleich, müssen auch die Hashcodes
     * gleich sein. Und die Gleichheit bestimmen eben equals()/compareTo().       <br><br>
     *
     * e.compareTo(null) sollte eine NullPointerException auslösen, auch wenn
     * e.equals(null) die Rückgabe false liefert.
     *
     * @param that
     *            Das mit diesem Objekt zu vergleichende andere Versions-Objekt.
     * @return <0 für this < that,                                                <br>
     *         =0 für this = that und                                             <br>
     *         >0 für this > that.                                                <br>
     *
     * @see equals
     * @see hashCode
     */
    @Override
    public int compareTo(Version that) {
        if (null == that) {
            throw new NullPointerException(
                    "Vergleich einer Version mit null ist unzulässig!");
        }

        /*
         * Zunächst testen wir, ob die String-Versionen gleich sind. In dem Fall
         * ist nichts weiter zu prüfen:
         */
        if (this.getVersion().equals(that.getVersion())) {
            return 0;
        }

        /*
         * Nun untersuchen wir die beiden Versionsnummerfolgen, bis zum Ende
         * des kürzeren:
         */
        int thisSize = this.numericVersion.size();
        int thatSize = that.numericVersion.size();
        int minSize = Math.max(thisSize, thatSize);
        for (int i=0; i<minSize; ++i) {
            int thisNumber = this.numericVersion.get(i);
            int thatNumber = that.numericVersion.get(i);
            if (thisNumber != thatNumber) {
                return thisNumber - thatNumber;
            }
        }

        /*
         * Kommt man hier an, sind beide bis zur minSize gleich, also hier
         * schauen, wer länger ist. Wer länger ist, wird als größer angesehen,
         * da 0.17.1.a größer als 0.17.1 angesehen wird:
         */

        if (thisSize == thatSize) {
            return 0;  // beide gleich lang und vorne gleich
        }
        if (thisSize == minSize) {
            return -1; // that ist länger und vorne gleich
        }
        else {
            return 1;  // this ist länger und vorne gleich
        }
    }

    /**
     * Erzeugt den Hashcode zu einem Versionsobjekt.
     *
     * @see equals
     * @see compareTo
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((numericVersion == null) ? 0 : numericVersion.hashCode());
        return result;
    }

    /**
     * Prüft auf die Gleichheit mit einem anderen Versionenobjekt. Dabei wird nur die numerische
     * Zahlenfolge der Version ausgewertet, nicht das Datum und nicht die Version als String.
     *
     * @see hashCode
     * @see compareTo
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        Version other = (Version) obj;
        if (numericVersion == null) {
            if (other.numericVersion != null) {
                return false;
            }
        }
        else if (!numericVersion.equals(other.numericVersion)) {
            return false;
        }
        return true;
    }

}
