package de.duehl.basics.text;

/*
 * Copyright 2019 Christian Dühl. All rights reserved.
 *
 * This program is free software. You can redistribute it and/or
 * modify it under the same terms as perl:
 *
 * general:  http://dev.perl.org/licenses/
 * GPL:      http://dev.perl.org/licenses/gpl1.html
 * artistic: http://dev.perl.org/licenses/artistic.html
 */

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.duehl.basics.debug.Assure;
import de.duehl.basics.text.Text;

/**
 * Diese Klasse enthält kleine, statische Hilfsmethoden für String, die Zahlen enthalten.
 *
 * @version 1.01     2019-04-26
 * @author Christian Dühl
 */

public class NumberString {

    /** Stellt eine Zahl mit Tausenderpunkten dar. */
    public static String taupu(int number) {
        DecimalFormat df = new DecimalFormat(",###");
        return df.format(number);
    }

    /** Stellt eine Zahl mit Tausenderpunkten dar. */
    public static String taupu(long number) {
        DecimalFormat df = new DecimalFormat(",###");
        return df.format(number);
    }

    /** Entfernt die Tausenderpunkte wieder aus dem String. */
    public static String removeTaupu(String taupu) {
        return taupu.replace(".", "");
    }

    private static final Pattern NUMBER_PATTERN = Pattern.compile("\\d+");
    private static final Pattern NUMBER_PATTERN_WITH_POINT = Pattern.compile("\\d[\\d.]*");
    private static final Pattern NUMBER_PATTERN_WITH_COMMA = Pattern.compile("\\d[\\d,]*");

    /**
     * Prüft, ob der gegebene String vollständig aus Ziffern besteht. Mindestens eine Ziffer muss
     * vorhanden sein.
     */
    public static boolean isDigitSequence(String text) {
        Matcher numberMatcher = NUMBER_PATTERN.matcher(text);
        return numberMatcher.matches();
    }

    /**
     * Prüft, ob der gegebene String vollständig aus Ziffern und Punkten besteht. Mindestens eine
     * Ziffer muss vorhanden sein und der String muss mit einer Ziffer beginnen.
     */
    public static boolean isDigitPerhapsWithPointSequence(String text) {
        Matcher numberMatcher = NUMBER_PATTERN_WITH_POINT.matcher(text);
        return numberMatcher.matches();
    }

    /**
     * Prüft, ob der gegebene String vollständig aus Ziffern und Punkten besteht. Mindestens eine
     * Ziffer muss vorhanden sein und der String muss mit einer Ziffer beginnen.
     */
    public static boolean isDigitPerhapsWithCommaSequence(String text) {
        Matcher numberMatcher = NUMBER_PATTERN_WITH_COMMA.matcher(text);
        return numberMatcher.matches();
    }

    /** Entfernt von einem String (der ein integer oder dergleichen darstellt) am Anfang Nullen. */
    public static String removeLeadingZeroes(String text) {
        String numberText = text;

        while (numberText.startsWith("0")) {
            numberText = numberText.substring(1);
        }

        return numberText;
    }

    /**
     * Entfernt von einem String (der ein double darstellt) am Ende Nullen und dann auch noch den
     * Dezimalpunkt.
     */
    public static String removeZeroEnding(String text) {
        String numberText = text;

        if (numberText.contains(".")) {
            while (numberText.endsWith("0")) {
                numberText = numberText.substring(0, numberText.length() - 1);
            }
        }
        if (numberText.endsWith(".")) {
            numberText = numberText.substring(0, numberText.length() - 1);
        }

        return numberText;
    }

    /**
     * Bereinigt Zahlenschreibweisen mit Kommas oder Punkten.
     *
     * @param source
     *            Zahlenstring mit Punkt oder Komma als Dezimaltrenner und Punkt
     *            als Tausendertrennzeichen.
     * @return Zahlenstring mit Punkt als Dezimaltrenner.
     */
    public static String cleanNumber(String source) {
        String cleanedSource = source;

        int numberOfPoints = Text.countPartInString(cleanedSource, ".");
        int numberOfKommas = Text.countPartInString(cleanedSource, ",");

        if (numberOfPoints > 1 || numberOfKommas > 0) {
            cleanedSource = cleanedSource.replace(".", "");
        }
        if (numberOfKommas > 1) {
            cleanedSource = cleanedSource.replace(",", "");
        }

        cleanedSource = cleanedSource.replace(",", ".");

        return cleanedSource;
    }

    private final static Pattern MATCHING_NUMBER_PATTERN = Pattern.compile("(\\d+)");

    /**
     * Findet alle ganzzahligen Integerzahlen im Eingabetext.
     *
     * @param text
     *            Eingabetext.
     * @return Liste mit den gefundenen Integerzahlen.
     */
    public static List<Integer> findNumbers(String text) {
        List<Integer> numbers = new ArrayList<>();
        Matcher matcher = MATCHING_NUMBER_PATTERN.matcher(text);
        while (matcher.find()) {
            String numberAsString = matcher.group(1);
            Integer number = Integer.parseInt(numberAsString);
            numbers.add(number);
        }

        return numbers;
    }

    /**
     * Wandelt eine Zahl in einen String um und fügt dabei führende Nullen hinzu. Aus number = 13
     * und numberOfDigits = 5 wird so "00013".
     *
     * @param number
     *            Umzuwandelnde Zahl.
     * @param numberOfDigits
     *            Anzahl der Stellen, entsprechend viele Nullen werden vorn aufgefüllt.
     * @return Zahl als String mit führenden Nullen.
     */
    public static String addLeadingZeroes(int number, int numberOfDigits) {
        String zeroNumber = Integer.toString(number);
        if (zeroNumber.length() >= numberOfDigits || number < 0 || numberOfDigits < 0) {
            return zeroNumber;
        }
        else {
            return Text.multipleString("0", numberOfDigits - zeroNumber.length()) + zeroNumber;
        }
    }

    /**
     * Wandelt eine Zahl in einen String um und fügt dabei führende Nullen hinzu. Aus number = 13
     * und numberOfDigits = 5 wird so "00013".
     *
     * @param number
     *            Umzuwandelnde Zahl.
     * @param numberOfDigits
     *            Anzahl der Stellen, entsprechend viele Nullen werden vorn aufgefüllt.
     * @return Zahl als String mit führenden Nullen.
     */
    public static String addLeadingZeroes(long number, int numberOfDigits) {
        String zeroNumber = Long.toString(number);
        if (zeroNumber.length() >= numberOfDigits || number < 0 || numberOfDigits < 0) {
            return zeroNumber;
        }
        else {
            return Text.multipleString("0", numberOfDigits - zeroNumber.length()) + zeroNumber;
        }
    }

    /**
     * Fügt vor einen String (mit Ziffern oder anderem) führende Nullen hinzu. Aus number = "13"
     * und numberOfDigits = 5 wird so "00013".
     *
     * @param number
     *            Umzuwandelnde Zahl oder anderes als String.
     * @param numberOfDigits
     *            Anzahl der Stellen, entsprechend viele Nullen werden vorn aufgefüllt.
     * @return Zahl als String mit führenden Nullen.
     */
    public static String addLeadingZeroes(String number, int numberOfDigits) {
        String zeroNumber = number;
        if (zeroNumber.length() >= numberOfDigits || number.startsWith("-") || numberOfDigits < 0) {
            return zeroNumber;
        }
        else {
            return Text.multipleString("0", numberOfDigits - zeroNumber.length()) + zeroNumber;
        }
    }

    /**
     * Die übergebende Zahl in einem String wird in einen entsprechenden int-Wert umgewandelt. Wenn
     * dabei (per Integer.parseInt(...)) eine NumberFormatException auftrat, wird statt dieser eine
     * RuntimeException mit dem übergebenen Fehlertext geworfen.
     *
     * @param numberString
     *            String mit einer Integerzahl darin (oder auch nicht).
     * @param errorMessage
     *            Fehlertext.
     * @return Gegebene Zahl als int.
     * @throws RuntimeException
     *             Falls ein Fehler beim Parsen der Zahl auftrat.
     */
    public static int parseInt(String numberString, String errorMessage) {
        try {
            return Integer.parseInt(numberString);
        }
        catch (NumberFormatException exception) {
            throw new RuntimeException(errorMessage, exception);
        }
    }

    /**
     * Die übergebende Zahl in einem String wird in einen entsprechenden int-Wert umgewandelt. Wenn
     * dabei (per Integer.parseInt(...)) eine NumberFormatException auftrat, wird statt dieser der
     * Fehlerwert zurückgegeben.
     *
     * @param numberString
     *            String mit einer Integerzahl darin (oder auch nicht).
     * @param failureValue
     *            Wert, der im Fehlerfall zurückgegeben wird.
     * @return Gegebene Zahl als int.
     */
    public static int parseIntIgnore(String numberString, int failureValue) {
        try {
            return Integer.parseInt(numberString);
        }
        catch (NumberFormatException exception) {
            return failureValue;
        }
    }

    /** Prüft, ob der übergebene String eine Integer-Zahl enthält. */
    public static boolean isInteger(String numberString) {
        try {
            Integer.parseInt(numberString);
            return true;
        }
        catch (NumberFormatException exception) {
            return false;
        }
    }

    /**
     * Die übergebende Zahl in einem String wird in einen entsprechenden Long-Wert umgewandelt.
     * Wenn dabei (per Long.parseLong(...)) eine NumberFormatException auftrat, wird statt dieser
     * eine RuntimeException mit dem übergebenen Fehlertext geworfen.
     *
     * @param numberString
     *            String mit einer Longzahl darin (oder auch nicht).
     * @param errorMessage
     *            Fehlertext.
     * @return Gegebene Zahl als long.
     * @throws RuntimeException
     *             Falls ein Fehler beim Parsen der Zahl auftrat.
     */
    public static long parseLong(String numberString, String errorMessage) {
        try {
            return Long.parseLong(numberString);
        }
        catch (NumberFormatException exception) {
            throw new RuntimeException(errorMessage, exception);
        }
    }

    /**
     * Die übergebende Zahl in einem String wird in einen entsprechenden Long-Wert umgewandelt.
     * Wenn dabei (per Long.parseLong(...)) eine NumberFormatException auftrat, wird statt dieser
     * der Fehlerwert zurückgegeben.
     *
     * @param numberString
     *            String mit einer Integerzahl darin (oder auch nicht).
     * @param failureValue
     *            Wert, der im Fehlerfall zurückgegeben wird.
     * @return Gegebene Zahl als long.
     */
    public static long parseLongIgnore(String numberString, long failureValue) {
        try {
            return Long.parseLong(numberString);
        }
        catch (NumberFormatException exception) {
            return failureValue;
        }
    }

    /**
     * Die übergebende Zahl in einem String wird in einen entsprechenden double-Wert umgewandelt.
     * Wenn dabei (per Double.parseDouble(...)) eine NumberFormatException auftrat, wird statt
     * dieser eine RuntimeException mit dem übergebenen Fehlertext geworfen.
     *
     * @param numberString
     *            String mit einer Double-Zahl darin (oder auch nicht).
     * @param errorMessage
     *            Fehlertext.
     * @return Gegebene Zahl als double.
     * @throws RuntimeException
     *             Falls ein Fehler beim Parsen der Zahl auftrat.
     */
    public static double parseDouble(String numberString, String errorMessage) {
        try {
            return Double.parseDouble(numberString);
        }
        catch (NumberFormatException exception) {
            throw new RuntimeException(errorMessage, exception);
        }
    }

    /**
     * Die übergebende Zahl in einem String wird in einen entsprechenden double-Wert umgewandelt.
     * Wenn dabei (per Double.parseDouble(...)) eine NumberFormatException auftrat, wird statt
     * dieser der Fehlerwert zurückgegeben.
     *
     * @param numberString
     *            String mit einer Double-Zahl darin (oder auch nicht).
     * @param failureValue
     *            Wert, der im Fehlerfall zurückgegeben wird.
     * @return Gegebene Zahl als double.
     */
    public static double parseDoubleIgnore(String numberString, double failureValue) {
        try {
            return Double.parseDouble(numberString);
        }
        catch (NumberFormatException exception) {
            return failureValue;
        }
    }

    /** Wandelt eine zweistellige Zahl in einem String in einen Integerwert um. */
    public static int twoDigitStringToInteger(String twoDigits) {
        if (twoDigits.length() != 2) {
            throw new RuntimeException(
                    "Es werden zwei Stellen erwartet! twoDigits = '" + twoDigits + "'.");
        }
        String number;
        if (twoDigits.startsWith("0")) {
            number = twoDigits.substring(1);
        }
        else {
            number = twoDigits;
        }
        return Integer.parseInt(number);
    }

    /** Gibt die Fließkommazahl als String mit zwei Nachkommastellen zurück. */
    public static String twoDecimalPlaces(double number) {
        return String.format("%.02f", number);
    }

    /**
     * Gibt die Fließkommazahl als String mit zwei Nachkommastellen zurück. Zusätzlich wird der
     * Teil vor dem Komma mit Tausenderpunkten versehen.
     */
    public static String twoDecimalPlacesTaupu(double number) {
        long before = (long) number;
        double after = number - before;
        String zeroCommaAndAfter = String.format("%.02f", after);
        String commaAndAfter = zeroCommaAndAfter.substring(1);
        return taupu(before) + commaAndAfter;
    }

    /** Bestimmt die Anzahl Stellen einer positiven ganzen Zahl. */
    public static int calculateDigits(int number) {
        Assure.isAtLeast(number, 0);

        // Vergleiche AnzahlStellenEinerZahl im ShortTestCodes.math
        return Integer.toString(number).length();
    }

    /**
     * Berechnet einen Prozentwert von count an der Gesamtzahl total und gibt ihn als String mit
     * zwei Nachkommastellen zurück.
     */
    public static String percent(int count, int total) {
        double percent = 100.0 * (count / (double) total);
        return String.format("%.2f", percent);
    }

    /** Gibt eine Zahl in einem String als sich selbst, "Null" oder "Eins" zurück. */
    public static String toGermanNumber(String number) {
        if (number.equals("0")) {
            return "Null";
        }
        else if (number.equals("1")) {
            return "Ein";
        }
        else {
            return number;
        }
    }

    /** Gibt abhängig von der Zahl die Plural- oder die Singular-Form des Wortes zurück. */
    public static String germanPlural(int number, String plural, String singular) {
        if (number == 1) {
            return singular;
        }
        else {
            return plural;
        }
    }

}
