package de.duehl.basics.text.xml;

/*
 * Copyright 2017 Christian Dühl. All rights reserved.
 *
 * This program is free software. You can redistribute it and/or
 * modify it under the same terms as perl:
 *
 * general:  http://dev.perl.org/licenses/
 * GPL:      http://dev.perl.org/licenses/gpl1.html
 * artistic: http://dev.perl.org/licenses/artistic.html
 */

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.duehl.basics.text.Text;

/**
 * Diese Klasse stellt Methoden zur Analyse und Extraktion aus einer XML-Datei eingelesenen Zeilen
 * zur Verfügung.
 *
 * @version 1.01     2017-07-20
 * @author Christian Dühl
 */

public class XmlAnalyser {

    private static final Pattern PARAMETER_PATTERN = Pattern.compile(
            "([-_A-Za-z0-9]+) *= *\"([^\"]+)\"");

    /**
     * Ermittelt alle Zeilen zwischen dem öffnenden und schließenden Tag, das öffnende und
     * schließende Tag inclusive.                                                          <br><br>
     *
     * Aus
     *     <foo>
     *         <bar parameter="7">
     *             Bar
     *         </bar>
     *     </foo>
     *
     * werden zum Tag "bar" die Zeilen
     *
     *         <bar parameter="7">
     *             Bar
     *         </bar>
     *
     * ermittelt.
     *
     * @param lines
     *            Zeilen in denen gesucht wird.
     * @param tag
     *            Tag nach dem gesucht wird.
     * @return Liste mit den relevanten Zeilen.
     */
    public static List<String> getLinesInTag(List<String> lines, String tag) {
        String start1 = "<" + tag + ">";
        String start2 = "<" + tag + " ";
        String end = "</" + tag + ">";

        List<String> tagLines = new ArrayList<>();
        boolean inTag = false;

        for (String line : lines) {
            if (line.contains(start1) || line.contains(start2)) {
                tagLines.add(line);
                inTag = true;
            }
            else if (line.contains(end)) {
                tagLines.add(line);
                inTag = false;
                break; // Um nicht spätere gleichnamige Tags zu erwischen. Wir nehmen nur den
                       // ersten Block!
            }
            else if (inTag) {
                tagLines.add(line);
            }
        }

        if (tagLines.isEmpty()) {
            throw new RuntimeException("In den eingelesenen Zeilen wurde kein Abschnitt mit dem "
                    + "Tag '" + tag + "' gefunden! Die Zeilen sind:\n\t"
                    + Text.joinWithLineBreak(lines));
        }

        return tagLines;
    }

    /**
     * Ermittelt aus Zeilen mit mehreren Blöcken des gleichen Tags die Listen von
     * Zeilen, die zu jeweils einem Block gehören.                                          <br><br>
     *
     * Aus
     *     <foo>
     *         <bar parameter="7">
     *             Bar1
     *         </bar>
     *         <bar>
     *             Bar2
     *         </bar>
     *     </foo>
     *
     * werden zum Tag "bar" die beiden Listen
     *
     *         <bar parameter="7">
     *             Bar1
     *         </bar>
     *
     * und
     *
     *         <bar>
     *             Bar2
     *         </bar>
     *
     * ermittelt.
     *
     * @param lines
     *            Zeilen in denen gesucht wird.
     * @param tag
     *            Tag nach dem gesucht wird.
     * @return Liste mit den Listen von jeweils zu einem gefundenen Tag gehörigen Zeilen.
     */
    public static List<List<String>> getMultipleLinesInTag(List<String> lines, String tag) {
        String start1 = "<" + tag + ">";
        String start2 = "<" + tag + " ";
        String end = "</" + tag + ">";

        List<List<String>> listOfTagLines = new ArrayList<>();
        List<String> tagLines = null;
        boolean inTag = false;

        for (String line : lines) {
            if (line.contains(start1) || line.contains(start2)) {
                tagLines = new ArrayList<>();
                tagLines.add(line);
                inTag = true;
            }
            else if (line.contains(end)) {
                tagLines.add(line);
                listOfTagLines.add(tagLines);
                tagLines = null;
                inTag = false;
            }
            else if (inTag) {
                tagLines.add(line);
            }
        }

        if (listOfTagLines.isEmpty()) {
            throw new RuntimeException("In den eingelesenen Zeilen wurde kein Abschnitt mit dem "
                    + "Tag '" + tag + "' gefunden! Die Zeilen sind:\n\t"
                    + Text.joinWithLineBreak(lines));
        }

        return listOfTagLines;
    }

    /**
     * Ermittelt die Zeile mit einem Tag in einer Zeile.                                    <br><br>
     *
     * Aus
     *     <foo>
     *         <bar parameter="7">Bar</bar>
     *     </foo>
     *
     * wird zum Tag "bar" die Zeile
     *
     *         <bar parameter="7">Bar</bar>
     *
     * ermittelt.
     *
     * @param lines
     *            Zeilen in denen gesucht wird.
     * @param tag
     *            Tag nach dem gesucht wird.
     * @return Gefundene Zeile.
     */
    public static String getFullTagInLine(List<String> lines, String tag) {
        String start1 = "<" + tag + ">";
        String start2 = "<" + tag + " ";
        String end = "</" + tag + ">";

        for (String line : lines) {
            if ((line.contains(start1) || line.contains(start2)) && line.contains(end)) {
                return line;
            }
        }

        throw new RuntimeException("In den eingelesenen Zeilen wurde keine Zeile mit dem "
                + "Tag '" + tag + "' gefunden! Die Zeilen sind:\n\t"
                    + Text.joinWithLineBreak(lines));
    }

    /**
     * Holt den Inhalt zwischen öffnendem und schließendem Tag heraus.                      <br><br>
     *
     * Aus
     *
     *         <bar parameter="7">Bar</bar>
     *
     * wird zum Tag "bar" der Inhalt "Bar" ermittelt.
     *
     * @param line
     *            Teile mit dem Tag.
     * @param tag
     *            Name des Tags.
     * @return Inhalt.
     */
    public static String getContentInTag(String line, String tag) {
        String start1 = "<" + tag + ">";
        String start2 = "<" + tag + " ";
        String end = "</" + tag + ">";

        int endPosition = line.indexOf(end);
        if (endPosition == -1) {
            throw new RuntimeException("Keinen schließenden Tag '" + end + "' in der Zeile '"
                    + line + "' gefunden!");
        }
        int startPosition1 = line.indexOf(start1);
        int startPosition2 = line.indexOf(start2);
        if (startPosition1 == -1 && startPosition2 == -1) {
            throw new RuntimeException("Keinen öffnenden Tag (weder '" + start1 + "'  noch '"
                    + start2 + "') in der Zeile '" + line + "' gefunden!");
        }
        int startPosition = 0;
        if (startPosition1 > 0) {
            startPosition = startPosition1 + start1.length();
        }
        else {
            startPosition = line.indexOf(">", startPosition2);
            if (startPosition == -1) {
                throw new RuntimeException("Es wurde kein '>' hinter dem öffnenden Tag '"
                        + start2 + "' in der Zeile '" + line + "' gefunden!");
            }
            ++startPosition;
        }

        String content = line.substring(startPosition, endPosition);
        return content;
    }

    /**
     * Holt den Inhalt zwischen öffnendem und schließendem Tag heraus.                      <br><br>
     *
     * Aus
     *     <foo>
     *         <bar parameter="7">Bar</bar>
     *     </foo>
     *
     * wird zum Tag "bar" der Inhalt "Bar" ermittelt.
     *
     * @param xml
     *            Zeilen aus einer XML-Datei.
     * @param tag
     *            Name des Tags.
     * @return Inhalt.
     */
    public static String getContentInFullTagLine(List<String> xml, String tag) {
        String line = XmlAnalyser.getFullTagInLine(xml, tag);
        String content = XmlAnalyser.getContentInTag(line, tag);
        return content;
    }

    /**
     * Holt den Inhalt zwischen öffnendem und schließendem Tag heraus.                      <br><br>
     *
     * Aus
     *     <foo>
     *         <bar parameter="7">Bar</bar>
     *     </foo>
     *
     * wird zum Tag "bar" der Inhalt "Bar" ermittelt.
     *
     * @param xml
     *            Zeilen aus einer XML-Datei.
     * @param tag
     *            Name des Tags.
     * @return Inhalt oder der leere String bei Fehlern.
     */
    public static String getContentInFullTagLineErrorToBlank(List<String> xml, String tag) {
        try {
            return getContentInFullTagLine(xml, tag);
        }
        catch (Exception exception) {
            return "";
        }
    }

    /**
     * Ermittelt die XML-Parameter aus einem öffnenden XML-Tag.
     *
     * @param line
     *            Zeile mit dem öffnenden XML-Tag.
     * @param tag
     *            Zu suchender Tag.
     * @return Liste mit den Parametern.
     */
    public static List<NamedXmlParameter> getParametersFromOpeningTag(String line, String tag) {
        String start2 = "<" + tag + " ";
        int startOfOpeningTagIndex = line.indexOf(start2);
        if (startOfOpeningTagIndex == -1) {
            throw new RuntimeException("Zum Tag '" + tag + "' wurd das öffneden Tag '" + start2
                    + "' nicht gefunden in der Zeile '" + line + "'!");
        }
        int endOfOpeningTagIndex = line.indexOf(">", startOfOpeningTagIndex);
        if (endOfOpeningTagIndex == -1) {
            throw new RuntimeException("Nach dem öffneden Tag '" + start2
                    + "' wurde kein '>' mehr gefunden in der Zeile '" + line + "'!");
        }

        //String openingTag = line.substring(startOfOpeningTagIndex, endOfOpeningTagIndex);
        String paramPart = line.substring(
                startOfOpeningTagIndex + 2 + tag.length(), // "<tag " kommt vorn noch weg.
                endOfOpeningTagIndex);                 // ">" kommt hinten noch weg.

        List<NamedXmlParameter> parameters = new ArrayList<>();

        Matcher matcher = PARAMETER_PATTERN.matcher(paramPart);
        while (matcher.find()) {
            String name = matcher.group(1);
            String content = matcher.group(2);
            NamedXmlParameter parameter = new NamedXmlParameter(name, content);
            parameters.add(parameter);
        }

        return parameters;
    }

    /**
     * Ermittelt die erste Zeile eines öffnenden Tags
     *
     * @param lines
     *            Zeilen in denen gesucht wird.
     * @param tag
     *            Zu suchender Tag.
     * @return Zeile mit dem öffnenden Tag.
     */
    public static String determineLineWithOpeningTag(List<String> lines, String tag) {
        for (String line : lines) {
            if (line.contains("<" + tag)) {
                return line;
            }
        }
        throw new RuntimeException("Es wurde kein öffnendes Tag '" + tag + "' in den übergebenen "
                + "Zeilen gefunden.\n\tlines = " + lines);
    }

    /**
     * Gibt an, ob es den öffnenden Tag in den übergebenen Zeilen gibt.
     *
     * @param lines
     *            Zeilen in denen gesucht wird.
     * @param tag
     *            Zu suchender Tag.
     * @return Zeile mit dem öffnenden Tag.
     */
    public static boolean existsOpeningTag(List<String> lines, String tag) {
        for (String line : lines) {
            if (line.contains("<" + tag)) {
                return true;
            }
        }
        return false;
    }

}
