package de.duehl.math.graph.ged.graph.io.ged3;

import static de.duehl.math.graph.ged.graph.io.ged3.Ged3LoadAndSaveConstants.*;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.duehl.basics.exceptions.WrongSourceFormatException;
import de.duehl.basics.text.NumberString;
import de.duehl.math.graph.ged.graph.GedColor;
import de.duehl.math.graph.ged.graph.Graph;
import de.duehl.math.graph.ged.graph.MetaData;
import de.duehl.math.graph.ged.graph.edge.Edge;
import de.duehl.math.graph.ged.graph.edge.EdgeStyle;
import de.duehl.math.graph.ged.graph.vertex.Vertex;
import de.duehl.math.graph.ged.pixel.PixelPoint;

/**
 * Diese Klasse lädt einen Graphen aus dem .ged3-Format.
 *
 * @version 1.01     2019-06-14
 * @author Christian Dühl
 */

public class Ged3FormatLoader {

    /** Der Graph in den hier geladen wird. */
    final private Graph graph;

    /** Enthält weitere Informationen zum Graphen. */
    final private MetaData meta; // zur Bequemlichkeit.

    /** Eingelesene Datei als ein String. */
    private String text;

    /** Aus dem Text extrahierter Teiltext mit den Daten zu den Ecken. */
    private String vertices;

    /** Aus dem Text extrahierter Teiltext mit den Daten zu den Kanten. */
    private String edges;

    /** Aus dem Text extrahierter Teiltext mit den Meta-Daten. */
    private String metaData;

    /** Anzahl der Ecken aus den Metadaten. */
    private int vertexNumber;

    /** Anzahl der Kanten aus den Metadaten. */
    private int edgeNumber;

    /**
     * Konstruktor.
     *
     * @param graph
     *            Der Graph in den hier geladen wird.
     */
    public Ged3FormatLoader(Graph graph) {
        this.graph = graph;
        meta = graph.getMeta();
    }

    /**
     * Lädt den Graphen aus dem .ged3-Format und ersetzt den aktuellen Graphen durch den geladenen.
     *
     * @param text
     *            Eingelesene Datei als ein String.
     * @throws WrongSourceFormatException
     *             Falls das Format nicht verstanden wird.
     * @throws NumberFormatException
     *             Falls beim Umwandeln einer Zahl ein Fehler auftritt.
     */
    public void loadFromSaveString(String text) {
        this.text = text;
        graph.clearGraph();

        determineTextParts();

        readMetaData();
        readVertices();
        readEdges();
    }

    private void determineTextParts() {
        determineVerticesPart();
        determineEdgesPart();
        determineMetaPart();
    }

    private static final Pattern VERTICES_PATTERN = Pattern.compile(
            ".*<" + VERTICES + ">((?:(?!</" + VERTICES + ">).)+)</" + VERTICES + ">.*",
            Pattern.DOTALL);

    private void determineVerticesPart() {
        Matcher verticesMatcher = VERTICES_PATTERN.matcher(text);
        if (verticesMatcher.matches()) {
            vertices = verticesMatcher.group(1);
        }
        else {
            throw new WrongSourceFormatException("Kann in der geladenen "
                    + "Datei keine Ecken (<" + VERTICES + "> ... </" + VERTICES + ">) "
                    + "finden!");
        }
    }

    private static final Pattern EDGES_PATTERN = Pattern.compile(
            ".*<" + EDGES + ">((?:(?!</" + EDGES + ">).)+)</" + EDGES + ">.*",
            Pattern.DOTALL);

    private void determineEdgesPart() {
        Matcher edgesMatcher = EDGES_PATTERN.matcher(text);
        if (edgesMatcher.matches()) {
            edges = edgesMatcher.group(1);
        }
        else {
            throw new WrongSourceFormatException("Kann in der geladenen "
                    + "Datei keine Kanten (<" + EDGES + "> ... </" + EDGES + ">) finden!");
        }
    }

    private static final Pattern META_DATA_PATTERN = Pattern.compile(
            ".*<" + GRAPH + ">(.*)<" + VERTICES + ">.*",
            Pattern.DOTALL);

    private void determineMetaPart() {
        Matcher getSimpleDataMatcher = META_DATA_PATTERN.matcher(text);
        if (getSimpleDataMatcher.matches()) {
            metaData = getSimpleDataMatcher.group(1);
        }
        else {
            throw new WrongSourceFormatException("Kann in der geladenen "
                    + "Datei den Bereich zwischen <" + GRAPH + "> und <" + VERTICES + ">"
                    + " nicht finden!");
        }
    }

    private static final Pattern FIND_TAGS_IN_METADATA_PATTERN = Pattern.compile(
            "<([a-z]+)>([^<>]+)</\\1>", Pattern.DOTALL);

    /** Liest die Metadaten ein. */
    private void readMetaData() {
        vertexNumber = -1;
        edgeNumber = -1;

        Matcher matcher = FIND_TAGS_IN_METADATA_PATTERN.matcher(metaData);
        while (matcher.find()) {
            String tag = matcher.group(1);
            String contents = matcher.group(2);
            switch (tag) {
                case META_VERTEX_NUMBER:
                    setMetaVertexNumber(contents);
                    break;
                case META_EDGE_NUMBER:
                    setMetaEdgeNumber(contents);
                    break;
                case META_DEFAULT_VERTEX_COLOR:
                    setMetaDefaultVertexColor(contents);
                    break;
                case META_DEFAULT_EDGE_COLOR:
                    setMetaDefaultEdgeColor(contents);
                    break;
                case META_DEFAULT_RADIUS:
                    setMetaDefaultVertexRadius(contents);
                    break;
                case META_BACKGROUND_COLOR:
                    setMetaBackgroundColor(contents);
                    break;
                case META_SHOW_GRID:
                    setMetaShowInGrid(contents);
                    break;
                case META_GRID_DISTANCE:
                    setMetaGridDistance(contents);
                    break;
                case META_SNAP_IN_GRID:
                    setMetaSnapInGrid(contents);
                    break;
                case META_GRID_COLOR:
                    setMetaGridColor(contents);
                    break;
                case META_BACKGROUND_PICTURE:
                    setMetaBackgoundPicture(contents);
                    break;
                case META_DEFAULT_EDGE_WIDTH:
                    setMetaDefaultEdgeWidth(contents);
                    break;
                case META_FILENAME:
                    setMetaFilename(contents);
                    break;
                default:
                    throw new WrongSourceFormatException(
                            "Unbekanntes Tag '" + tag + "' in den Metadaten gefunden!");
            }
        }
    }

    private void setMetaVertexNumber(String vertexNumberAsString) {
        vertexNumber = Integer.parseInt(vertexNumberAsString);
    }

    private void setMetaEdgeNumber(String edgeNumberAsString) {
        edgeNumber = Integer.parseInt(edgeNumberAsString);
    }

    private void setMetaDefaultVertexColor(String defaultVertexColorAsString) {
        GedColor defaultVertexColor = new GedColor(defaultVertexColorAsString);
        meta.setDefaultVertexColor(defaultVertexColor);
    }

    private void setMetaDefaultEdgeColor(String defaultEdgeColorAsString) {
        GedColor defaultEdgeColor = new GedColor(defaultEdgeColorAsString);
        meta.setDefaultEdgeColor(defaultEdgeColor);
    }

    private void setMetaDefaultVertexRadius(String defaultRadiusAsString) {
        int defaultRadius = Integer.parseInt(defaultRadiusAsString);
        meta.setDefaultRadius(defaultRadius);
    }

    private void setMetaBackgroundColor(String backgroundColorAsString) {
        GedColor backgroundColor = new GedColor(backgroundColorAsString);
        meta.setBackgroundColor(backgroundColor);
    }

    private void setMetaShowInGrid(String showGridAsString) {
        int showGridAsInt = Integer.parseInt(showGridAsString);
        boolean showGrid = showGridAsInt > 0;
        meta.setShowGrid(showGrid);
    }

    private void setMetaGridDistance(String gridDistanceAsString) {
        int gridDistance = Integer.parseInt(gridDistanceAsString);
        meta.setGridDistance(gridDistance);
    }

    private void setMetaSnapInGrid(String snapInGridAsString) {
        int snapInGridAsInt = Integer.parseInt(snapInGridAsString);
        boolean snapInGrid = snapInGridAsInt > 0;
        meta.setSnapInGrid(snapInGrid);
    }

    private void setMetaGridColor(String gridColorAsString) {
        GedColor gridColor = new GedColor(gridColorAsString);
        meta.setGridColor(gridColor);
    }

    private void setMetaBackgoundPicture(String backgroundPicture) {
        meta.setBackgroundPicture(backgroundPicture);
    }

    private void setMetaDefaultEdgeWidth(String defaultEdgeWidthAsString) {
        int defaultEdgeWidth = Integer.parseInt(defaultEdgeWidthAsString);
        meta.setDefaultEdgeWidth(defaultEdgeWidth);
    }

    @SuppressWarnings("unused")
    private void setMetaFilename(String filename) {
        // TODO Wenn man den doch mal speichern will, ist hier die richtige Stelle.
    }

    private static final Pattern VERTEX_PATTERN = Pattern.compile(
            "<" + VERTEX + ">((?:(?!</" + VERTEX + ">).)+)</" + VERTEX + ">", Pattern.DOTALL);
    private static final Pattern INNER_VERTEX_PATTERN = Pattern.compile(
            "<([a-z]+)>([^<>]+)</\\1>");

    /** Eckendaten ermitteln. */
    private void readVertices() {
        int countVertices = 0;

        Matcher matcher = VERTEX_PATTERN.matcher(vertices);
        while (matcher.find()) {
            ++countVertices;
            String contents = matcher.group(1);
            readVertex(contents);
        }

        if (countVertices != vertexNumber) {
            throw new WrongSourceFormatException("Anzahl der ausgelesenen Ecken (" + countVertices
                    + ") stimmt nicht mit der Eckennummer aus der Datei (" + vertexNumber
                    + ") überein!");
        }
    }

    private void readVertex(String contents) {
        Matcher innerMatcher = INNER_VERTEX_PATTERN.matcher(contents);
        Vertex vertex = new Vertex();
        while (innerMatcher.find()) {
            String tag = innerMatcher.group(1);
            String innerContents = innerMatcher.group(2);
            //System.out.println("tag: '" + tag + "', innerContents: '" + innerContents + "'");
            switch (tag) {
                case VERTEX_X:
                    setVertexX(vertex, innerContents);
                    break;
                case VERTEX_Y:
                    setVertexY(vertex, innerContents);
                    break;
                case VERTEX_COLOR:
                    if (hasDeefaultValue(innerContents)) {
                        vertex.setColor(GedColor.DEFAULT);
                    }
                    else {
                        GedColor color = new GedColor(innerContents);
                        vertex.setColor(color);
                    }
                    break;
                case VERTEX_RADIUS:
                    if (hasDeefaultValue(innerContents)) {
                        vertex.setRadiusToDefault();
                    }
                    else {
                        int radius = Integer.parseInt(innerContents);
                        vertex.setRadius(radius);
                    }
                    break;
                case VERTEX_WEIGHT:
                    if (hasDeefaultValue(innerContents)) {
                        vertex.setWeight(Vertex.NO_WEIGHT);
                    }
                    else {
                        double weight = Double
                                .parseDouble(innerContents);
                        vertex.setWeight(weight);
                    }
                    break;
                case VERTEX_TITLE:
                    vertex.setLabel(innerContents);
                    break;
                case VERTEX_TITLE_POSITION: {
                    PixelPoint distance = calculateDistanceFromOrientation(vertex, innerContents,
                            vertex.getLabel());
                    vertex.setLabelDistance(distance);
                    break;
                }
                case VERTEX_WEIGHT_POSITION: {
                    PixelPoint distance = calculateDistanceFromOrientation(vertex, innerContents,
                            Double.toString(vertex.getWeight()));
                    vertex.setWeightDistance(distance);
                    break;
                }
                case VERTEX_TITLE_COLOR:
                    if (hasDeefaultValue(innerContents)) {
                        vertex.setLabelColor(GedColor.DEFAULT);
                    }
                    else {
                        GedColor color = new GedColor(innerContents);
                        vertex.setLabelColor(color);
                    }
                    break;
                case VERTEX_WEIGHT_COLOR:
                    if (hasDeefaultValue(innerContents)) {
                        vertex.setWeightColor(GedColor.DEFAULT);
                    }
                    else {
                        GedColor color = new GedColor(innerContents);
                        vertex.setWeightColor(color);
                    }
                    break;
            }
        }
        graph.addVertex(vertex);
    }

    private void setVertexX(Vertex vertex, String innerContents) {
        int x = Integer.parseInt(innerContents);
        vertex.setX(x);
    }

    private void setVertexY(Vertex vertex, String innerContents) {
        int y = Integer.parseInt(innerContents);
        vertex.setY(y);
    }

    /** Berechnet Abstandskoordinaten aus einer Orientierungsangabe wie "SW". */
    private PixelPoint calculateDistanceFromOrientation(Vertex vertex, String orientation,
            String text) {
        int radius = vertex.getRadius();
        return calculateDistanceFromOrientation(radius, orientation, text);
    }

    /** Berechnet Abstandskoordinaten aus einer Orientierungsangabe wie "SW". */
    private PixelPoint calculateDistanceFromOrientation(int radius, String orientation,
            String text) {
        /*
         * Aus der Dokumentation zu drawString:
         *     "The baseline of the leftmost character is at position (x, y) in this graphics
         *     context's coordinate system."
         * Wir können hier also nicht mit Ankern arbeiten wie unter Perl/Tk, was die Sache sehr
         * viel einfacher machen würde.
         */
        PixelPoint distance = new PixelPoint();

        String removedText = NumberString.removeZeroEnding(text);

        /* Geschätzte Länge des darzustellenden Textes in Pixeln: */
        int textLength = removedText.length() * 8;

        /* geschätzte Höhe in Pixeln: */
        int textHeight = 10;

        int delta;
        if (radius == Vertex.USE_DEFAULT_RADIUS) {
            delta = 1 + 2 * meta.getDefaultRadius();
        }
        else {
            delta = 1 + 2 * radius;
        }

        switch (orientation) {
            case POSITION_N:
                distance.setX(-0.5 * textLength);
                distance.setY(-delta);
                break;
            case "NE":
            case POSITION_NO:
                distance.setX(delta);
                distance.setY(-delta);
                break;
            case "E":
            case POSITION_O:
                distance.setX(delta);
                distance.setY(0.5 * textHeight);
                break;
            case "SE":
            case POSITION_SO:
                distance.setX(delta);
                distance.setY(delta + textHeight);
                break;
            case POSITION_S:
                distance.setX(-0.5 * textLength);
                distance.setY(delta + textHeight);
                break;
            case POSITION_SW:
                distance.setX(-textLength - delta);
                distance.setY(delta + textHeight);
                break;
            case POSITION_W:
                distance.setX(-textLength - delta);
                distance.setY(0.5 * textHeight);
                break;
            case POSITION_NW:
                distance.setX(-textLength - delta);
                distance.setY(-delta);
                break;
            default:
                distance.setX(0);
                distance.setY(0);
                break;
        }
        return distance;
    }

    private static final Pattern EDGE_PATTERN = Pattern.compile(
            "<" + EDGE + ">((?:(?!</" + EDGE + ">).)+)</" + EDGE + ">", Pattern.DOTALL);

    /** Liest die Kantendaten ein. */
    private void readEdges() {
        int countEdges = 0;

        Matcher matcher = EDGE_PATTERN.matcher(edges);
        while (matcher.find()) {
            ++countEdges;
            readEdge(matcher.group(1));
        }
        if (countEdges != edgeNumber) {
            throw new WrongSourceFormatException("Anzahl der ausgelesenen Kanten (" + countEdges
                    + ") stimmt nicht mit der Kantennummer aus der Datei (" + edgeNumber
                    + ") überein!");
        }
    }

    private static final Pattern INNER_EDGE_PATTERN = Pattern.compile(
            "<([a-z]+)>([^<>]+)</\\1>");

    private void readEdge(String contents) {
        Matcher innerMatcher = INNER_EDGE_PATTERN.matcher(contents);
        String startVertexIndexString = null;
        String targetVertexIndexString = null;
        GedColor edgeColor = GedColor.DEFAULT;
        boolean forwardEdge = false;
        boolean backwardEdge = false;
        EdgeStyle style = EdgeStyle.NORMAL;
        int width = 1;
        String label = "";
        PixelPoint labelDistance = new PixelPoint();
        GedColor labelColor = GedColor.DEFAULT;
        double weight = 0.0;
        PixelPoint weightDistance = new PixelPoint();
        GedColor weightColor = GedColor.DEFAULT;
        while (innerMatcher.find()) {
            String tag = innerMatcher.group(1);
            String innerContents = innerMatcher.group(2);
            //System.out.println("tag: '" + tag + "', contents: '" + innerContents + "'");
            switch (tag) {
                case EDGE_START_VERTEX:
                    startVertexIndexString = innerContents;
                    break;
                case EDGE_END_VERTEX:
                    targetVertexIndexString = innerContents;
                    break;
                case EDGE_COLOR:
                    if (hasDeefaultValue(innerContents)) {
                        edgeColor = GedColor.DEFAULT;
                    }
                    else {
                        edgeColor = new GedColor(innerContents);
                    }
                    break;
                case EDGE_FORM: //  ---, --->, <--- oder <---> Achtung!
                    /*
                     *  k_ins => 'normale Kante einfügen',
                     *  k_inh => 'Kante einfügen: Pfeil hin',
                     *  k_inz => 'Kante einfügen: Pfeil zurück',
                     *  k_inb => 'Kante einfügen: Pfeil hin und zurück'
                     */
                    switch (innerContents) {
                        case "k_ins":
                        case "k_inb":
                            forwardEdge = true;
                            backwardEdge = true;
                            break;
                        case "k_inh":
                            forwardEdge = true;
                            break;
                        case "k_inz":
                            backwardEdge = true;
                            break;
                    }
                    break;
                case EDGE_TYPE:
                    switch (innerContents) {
                        case EDGE_TYPE_NORMAL:
                            style = EdgeStyle.NORMAL;
                            break;
                        case EDGE_TYPE_GESTRICHELT:
                            style = EdgeStyle.DASHED;
                            break;
                        case EDGE_TYPE_GEPUNKTET:
                            style = EdgeStyle.DOTTED;
                            break;
                    }
                    break;
                case EDGE_WIDTH:
                    if (hasDeefaultValue(innerContents)) {
                        width = Edge.USE_DEFAULT_WIDTH;
                    }
                    else {
                        width = Integer.parseInt(innerContents);
                    }
                    break;
                case EDGE_TITLE:
                    label = innerContents;
                    break;
                case EDGE_TITLE_POSITION:
                    labelDistance = calculateDistanceFromOrientation(meta.getDefaultRadius(),
                            innerContents, label);
                    break;
                case EDGE_TITLE_COLOR:
                    if (hasDeefaultValue(innerContents)) {
                        labelColor = GedColor.DEFAULT;
                    }
                    else {
                        labelColor = new GedColor(innerContents);
                    }
                    break;
                case EDGE_WEIGHT:
                    if (hasDeefaultValue(innerContents)) {
                        weight = Edge.NO_WEIGHT;
                    }
                    else {
                        weight = Double.parseDouble(innerContents);
                    }
                    break;
                case EDGE_WEIGHT_POSITION:
                    weightDistance = calculateDistanceFromOrientation(
                            meta.getDefaultRadius(), innerContents,
                            NumberString.removeZeroEnding(Double.toString(weight)));
                    break;
                case EDGE_WEIGHT_COLOR:
                    if (hasDeefaultValue(innerContents)) {
                        weightColor = GedColor.DEFAULT;
                    }
                    else {
                        weightColor = new GedColor(innerContents);
                    }
                    break;
                case EDGE_NUMBER:
                    // wird hier anders verwaltet, keine Nummer nötig.
                    break;
            }
        }
        if (forwardEdge || backwardEdge) {
            int startVertexIndex = Integer.parseInt(startVertexIndexString);
            int targetVertexIndex = Integer.parseInt(targetVertexIndexString);
            Edge edge =  new Edge(-1, -1);
            edge.setColor(edgeColor);
            edge.setStyle(style);
            edge.setWidth(width);
            edge.setLabel(label);
            edge.setLabelColor(labelColor);
            edge.setLabelDistance(labelDistance);
            edge.setWeight(weight);
            edge.setWeightColor(weightColor);
            edge.setWeightDistance(weightDistance);
            if (forwardEdge) {
                edge.setStartVertexIndex(startVertexIndex);
                edge.setTargetVertexIndex(targetVertexIndex);
                graph.addEdge(edge);
            }
            if (backwardEdge) {
                edge.setStartVertexIndex(targetVertexIndex);
                edge.setTargetVertexIndex(startVertexIndex);
                graph.addEdge(edge);
            }
        }
        else {
            throw new RuntimeException("Kante zwischen den Ecken " + startVertexIndexString
                    + " und " + targetVertexIndexString + " ist keine Form definiert.");
        }
    }

    private boolean hasDeefaultValue(String innerContents) {
        return "undef".equals(innerContents) || "default".equals(innerContents);
    }

}
