package de.duehl.math.graph.ged.ui.graphpanel;

import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionAdapter;
import java.util.ArrayList;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.JPanel;

import de.duehl.math.geometry.Circle;
import de.duehl.math.geometry.Geometry;
import de.duehl.math.geometry.Line;
import de.duehl.math.geometry.Point;
import de.duehl.math.graph.ged.graph.GedColor;
import de.duehl.math.graph.ged.graph.Graph;
import de.duehl.math.graph.ged.graph.edge.Edge;
import de.duehl.math.graph.ged.graph.edge.EdgeStyle;
import de.duehl.math.graph.ged.graph.edge.Edges;
import de.duehl.math.graph.ged.graph.vertex.Vertex;
import de.duehl.math.graph.ged.logic.Logic;
import de.duehl.math.graph.ged.logic.ObserverPattern.Observable;
import de.duehl.math.graph.ged.logic.ObserverPattern.Observer;
import de.duehl.math.graph.ged.pixel.PixelPoint;
import de.duehl.math.graph.ged.resources.IconDefinitions;
import de.duehl.math.graph.ged.ui.GuiElements;
import de.duehl.math.graph.ged.ui.dialogs.GraphErrorDialog;
import de.duehl.swing.ui.GuiTools;
import de.duehl.swing.ui.dialogs.ErrorDialog;
import de.duehl.swing.ui.resources.IconLoader;
import de.duehl.basics.logging.Logger;

/**
 * Diese Klasse steht für das JPanel, in das der Graph eingezeichnet wird.
 *
 * @version 1.01     2019-07-02
 * @author Christian Dühl
 */

public class GraphPanel extends JPanel implements MouseListener, Observable {

    private static final long serialVersionUID = 3791742371384588734L;

    /** Bearbeitungsmodus */
    private EditModus editModus;

    /* Variablen für das Bewegen der Ecken (MOVE_VERTEX): */

    /** Gibt an, ob der Kreis gerade mit der Maus bewegt werden soll. */
    private boolean circleSticksAtMouse;

    /** Ecke, die bewegt wird. */
    private Vertex stickyVertex;

    /* Variablen für das Einfügen von Kanten nach: */
    private Vertex firstEdgeVertex;

    private boolean edgeToVertexSticksAtMouse;

    /* Variablen für das Einfügen von Kanten nach: */
    /* firstEdgeVertex wird mitbenutzt */

    private boolean doubleEdgeSticksAtMouse;

    /* Sonstige Variablen: */

    /** X-Koordinate der aktuellen Mausposition. */
    private int mouseActualX;

    /** Y-Koordinate der aktuellen Mausposition. */
    private int mouseActualY;

    /** Graph, der dargestellt wird. */
    private Graph graph;

    /** Objekt, das den Graphen zeichnet. */
    private GraphPainter graphPainter;

    /** Zugriff auf die Programmlogik */
    private Logic logic;

    /** Der Logger. */
    private Logger logger;

    /** Dateiname des Hintergrundbildes zur Pufferung. */
    private String savedPicture;

    /** Eingelesenes Hintergrundbild zur Pufferung. */
    private Image savedImage;

    private GuiElements elements;

    /**
     * Konstruktor
     *
     * @param width
     *            Breite des Spielpanels.
     * @param height
     *            Höhe des Spielpanels.
     * @param logic
     */
    public GraphPanel(int width, int height, Logic logic, GuiElements elements) {
        graph = new Graph();
        graphPainter = new GraphPainter();
        this.logic = logic;
        this.elements = elements;
        logger = logic.getLogger();
        savedPicture = "";
        savedImage = null;
        init();
    }

    /**
     * Initialisiert die Werte (zu Beginn und wenn ein neuer Graph geladen
     * wird).
     */
    private void init() {
        circleSticksAtMouse = false;
        edgeToVertexSticksAtMouse = false;
        doubleEdgeSticksAtMouse = false;
        stickyVertex = null;
        firstEdgeVertex = null;
    }

    /** Setzt alles zurück, was zum verschieben oder einfügen dient. */
    public void interruptEditing() {
        init();
        repaint();
    }

    /**
     * Erzeugt den MouseMotionListener, der auf Variablen dieser Klasse
     * zugreift.
     */
    public void doAddMouseMotionListener() {
        this.addMouseMotionListener(new MouseMotionAdapter() {
            @Override
            public void mouseMoved(MouseEvent event) {
                workOnMouseChanges(event);
            }
            @Override
            public void mouseDragged(MouseEvent event) {
                workOnMouseChanges(event);
            }
            private void workOnMouseChanges(MouseEvent event) {
                try {
                    tryWorkOnMouseChanges(event);
                }
                catch (Exception exception) {
                    error(exception);
                }
            }
            private void tryWorkOnMouseChanges(MouseEvent event) {
                mouseActualX = event.getX();
                mouseActualY = event.getY();
                if (circleSticksAtMouse) {
                    setVertexKoordinates(mouseActualX, mouseActualY, stickyVertex);
                }
                repaint();
            }
        });
    }

    /** Auf Panel zeichnen. Fehlerbehandlung. */
    @Override
    public void paintComponent(Graphics graphics) {
        super.paintComponent(graphics);

        try {
            paintGraphPanel(graphics);
        }
        catch (Exception exception) {
            error(exception);
        }
    }

    /** Auf Panel zeichnen. */
    private void paintGraphPanel(Graphics graphics) {
        paintBackground(graphics);

        /*
        Graphics2D g2 = (Graphics2D) graphics;
        g2.translate(0, getHeight());
        g2.scale(1.0, -1.0);
        */

        /*
        graphics.setColor(Color.BLUE);
        graphics.drawString("Mouse position: " + mouseActualX + ", "
                + mouseActualY, 10, 20);
        */

        /* Größe des GraphPanel anzeigen: */
        showGraphPanleSize();

        /* Gitter darstellen: */
        if (graph.getMeta().isShowGrid()) {
            paintGrid(graphics);
        }

        /* Graphen Darstellen: */
        graphPainter.paintGraph(graphics, graph);

        /* Wenn gerade Kanten erzeugt werden, diese darstellen: */
        if (edgeToVertexSticksAtMouse || doubleEdgeSticksAtMouse) {
            paintDrawingEdges();
        }
    }

    private void showGraphPanleSize() {
        int width = getWidth();
        int height = getHeight();
        logic.showGraphPanelSize(width, height);
    }

    private void paintGrid(Graphics graphics) {
        GedColor gridColor = graph.getMeta().getGridColor();
        graphics.setColor(gridColor.toSwingColor());

        int width = getWidth();
        int height = getHeight();
        int distance = (int) Math.round(graph.getMeta().getGridDistance());
        for (int x = distance; x < width; x += distance) {
            graphics.drawLine(x, 0, x, height);
        }
        for (int y = distance; y < width; y += distance) {
            graphics.drawLine(0, y, width, y);
        }
    }

    private void paintDrawingEdges() {
        PixelPoint start = firstEdgeVertex.getCoordinates();
        PixelPoint end = new PixelPoint(mouseActualX, mouseActualY);
        Line line = new Line(start.toMathPoint(), end.toMathPoint());

        int startRadius = firstEdgeVertex.getRadius();
        if (Vertex.USE_DEFAULT_RADIUS == startRadius) {
            startRadius = graph.getMeta().getDefaultRadius();
        }
        Geometry geometry = new Geometry();
        Point realStart = geometry.getPointOnLineWithDistanceToStartPoint(line, startRadius);
        Line realLine = new Line(realStart, end.toMathPoint());

        /* Kante zeichnen: */
        GedColor edgeColor = graph.getMeta().getDefaultEdgeColor();
        EdgeStyle style = EdgeStyle.NORMAL;
        int widthWhileDrawing = 1;
        graphPainter.drawLine(realLine, edgeColor, style, widthWhileDrawing);
        /* Spitze einzeichnen: */
        if (graph.isDirected()) {
            graphPainter.paintArrow(realLine, edgeColor, style, widthWhileDrawing);
        }

        /*
         * Falls zwei Kanten hinzugefügt werden sollen, dann muss auch die
         * andere Spitze gezeichnet werden:
         */
        if (doubleEdgeSticksAtMouse) {
            realLine.swap();
            graphPainter.paintArrow(realLine, edgeColor, style, widthWhileDrawing);
        }
    }

    /** Stellt den Hintergrund dar - Farbe oder Bild. */
    private void paintBackground(Graphics graphics) {
        paintBackgroundColor();
        paintBackgroundPicture(graphics);
    }

    private void paintBackgroundColor() {
        GedColor backgroundColor = graph.getMeta().getBackgroundColor();
        setBackground(backgroundColor.toSwingColor());
    }

    private void paintBackgroundPicture(Graphics graphics) {
        readBackgroundPictureFromDiskIfItHasChanged();

        if (!savedPicture.isEmpty()) {
            graphics.drawImage(savedImage, 0, 0, null);
        }
    }

    private void readBackgroundPictureFromDiskIfItHasChanged() {
        String picture = graph.getMeta().getBackgroundPicture();
        if (!picture.equals(savedPicture)) {
            savedPicture = picture;
            if (picture.isEmpty()) {
                savedImage = null;
            }
            else {
                Image img = new ImageIcon(picture).getImage();
                savedImage = img;
            }
        }
    }

    /** Überschriebene Methode des MouseListeners */
    @Override
    public void mouseClicked(MouseEvent event) {
    }

    /** Überschriebene Methode des MouseListeners */
    @Override
    public void mouseEntered(MouseEvent event) {
    }

    /** Überschriebene Methode des MouseListeners */
    @Override
    public void mouseExited(MouseEvent event) {
        try {
            tryToReactAtMouseExited(event);
        }
        catch (Exception exception) {
            error(exception);
        }
    }

    /** Reagiert auf das Verlassen des Graphenpanels mit der Maus. */
    private void tryToReactAtMouseExited(MouseEvent event) {
        if (editModus == EditModus.MOVE_VERTEX && circleSticksAtMouse) {
            releaseCircle(event);
        }
    }

    /** Überschriebene Methode des MouseListeners */
    @Override
    public void mousePressed(MouseEvent event) {
        try {
            tryToReactAtMousePressed(event);
        }
        catch (Exception exception) {
            error(exception);
        }
    }

    /** Reagiert auf das Drücken eines Mausknopfes. */
    private void tryToReactAtMousePressed(MouseEvent event) {
        PixelPoint coords = new PixelPoint(event.getPoint());
        Vertex vertex = getVertexAt(coords);
        if (event.getButton() == MouseEvent.BUTTON1) {
            if (editModus == EditModus.MOVE_VERTEX && !circleSticksAtMouse) {
                if (null != vertex) {
                    circleSticksAtMouse = true;
                    stickyVertex = vertex;
                }
            }
        }
    }

    /** Überschriebene Methode des MouseListeners */
    @Override
    public void mouseReleased(MouseEvent event) {
        try {
            tryToReactAtMouseReleased(event);
        }
        catch (Exception exception) {
            error(exception);
        }
    }

    /** Zeigt einen Fehler an. */
    private void error(Throwable exception) {
        String message = "Es ist ein nicht abgefangener Fehler aufgetreten!";
        IconDefinitions iconDefinitions = new IconDefinitions();
        IconLoader iconLoader = iconDefinitions.createIconLoader();
        Image programImage = iconLoader.loadProgramIconImage();
        ErrorDialog dialog = new GraphErrorDialog(message, exception, graph,
                programImage, logger.getLogFileName(), elements.getFrameLocation());
        dialog.setVisible(true);
    }

    /** Reagiert auf das Loslassen eines Mausknopfes. */
    private void tryToReactAtMouseReleased(MouseEvent event) {
        PixelPoint coords = new PixelPoint(event.getPoint());
        Vertex vertex = getVertexAt(coords);
        if (event.getButton() == MouseEvent.BUTTON1) {
            switch (editModus) {
                case ADD_VERTEX:
                    /* Es soll keine andere Ecke in der Nähe sein! */
                    if (!areThereVerticesAt(coords)) {
                        addVertex(coords, Vertex.USE_DEFAULT_RADIUS, GedColor.DEFAULT);
                    }
                    //if (modus == Modus.ADD_VERTEX) throw new RuntimeException("Au weia!!!");
                    break;
                case ADD_EDGE_TO_VERTEX:
                case ADD_UNDIRECTED_EDGE:
                    if (null == vertex) {
                        beep();
                        break;
                    }
                    if (firstEdgeVertex == null) {
                        edgeToVertexSticksAtMouse = true;
                        firstEdgeVertex = vertex;
                    }
                    /* Keine Schlingen! */
                    else if (vertex.equals(firstEdgeVertex)) {
                        beep();
                    }
                    else {
                        addEdge(firstEdgeVertex, vertex, GedColor.DEFAULT);
                        edgeToVertexSticksAtMouse = false;
                        firstEdgeVertex = null;
                    }
                    break;
                case ADD_BOTH_DIRECTED_EDGE:
                    if (null == vertex) {
                        beep();
                        break;
                    }
                    if (firstEdgeVertex == null) {
                        doubleEdgeSticksAtMouse = true;
                        firstEdgeVertex = vertex;
                    }
                    /* Keine Schlingen! */
                    else if (vertex.equals(firstEdgeVertex)) {
                        beep();
                    }
                    else {
                        /* Kante hin einfügen: */
                        addEdge(firstEdgeVertex, vertex, GedColor.DEFAULT);

                        /* Kante zurück einfügen: */
                        addEdge(vertex, firstEdgeVertex, GedColor.DEFAULT);

                        doubleEdgeSticksAtMouse = false;
                        firstEdgeVertex = null;
                    }
                    break;
                case MOVE_VERTEX:
                    if (circleSticksAtMouse) {
                        releaseCircle(event);
                    }
                    break;
                case DELETE_VERTEX:
                    if (null != vertex) {
                        graph.removeVertex(vertex);
                        setChange();
                    }
                    break;
                case MERGE_VERTEX:
                    if (null != vertex) {
                        graph.mergeVertex(vertex);
                        setChange();
                    }
                    break;
                case DELETE_EDGE:
                    deleteEdge(coords, false);
                    break;
                case DELETE_MULTIPLE_EDGES:
                    deleteEdge(coords, true);
                    break;
                case INSPECT:
                    inspect(coords, vertex);
                    break;
                default:
                    break;
            }
        }
        /* Rechtsklick: */
        else if (event.getButton() == MouseEvent.BUTTON3) {
            inspect(coords, vertex);
        }
    }

    /**
     * Kontextmenü für Ecken oder Kanten.
     *
     * @param coords
     *            Koordinaten des Mausklicks.
     * @param vertex
     *            Ermittelte Ecke.
     */
    private void inspect(PixelPoint coords, Vertex vertex) {
        if (null != vertex) {
            Vertex tempVertex = new Vertex(vertex);
            logic.showVertexPopupMenu(vertex, coords);
            if (!vertex.equals(tempVertex)) {
                setChange();
                repaint();
            }
        }
        else {
            Edges edgesAtCoord = getAllEdgesAt(coords);
            if (edgesAtCoord.size() == 1) {
                Edge edge = edgesAtCoord.get(0);
                Edge tempEdge = new Edge(edge);
                logic.showEdgePopupMenu(edge, coords);
                if (!edge.equals(tempEdge)) {
                    setChange();
                    repaint();
                }
            }
        }
    }

    /**
     * Löscht eine oder alle Kanten, die nah genug an den gegebenen Koordinaten sind.
     *
     * @param coords
     *            Koordinaten
     * @param multipleDelete
     *            Gibt an, ob alle dort gefundenen Kanten entfernt werden soll, oder nur eine (die
     *            dann aber eindeutig sein muss).
     */
    private void deleteEdge(PixelPoint coords, boolean multipleDelete) {
        Edges edgesAtCoords = getAllEdgesAt(coords);
        if (1 == edgesAtCoords.size() || multipleDelete) {
            deleteAllEdgesFrom(edgesAtCoords);
            setChange();
        }
        /*
         * Anderenfalls haben wir mehrere Kanten gefunden, aber wollen nur eine löschen. Nun fragen
         * wir die Logik, welche Kanten davon wir löschen sollen:
         */
        else if (!edgesAtCoords.isEmpty()) {
            Edges edgesToDelete = logic.determineEdgesToDelete(edgesAtCoords, coords);
            if (!edgesToDelete.isEmpty()) {
                deleteAllEdgesFrom(edgesToDelete);
                setChange();
            }
        }
    }

    /**
     * Entfernt alle Kanten aus dem Graphen, die in der übergebenen Datenstruktur enthalten sind.
     *
     * @param edgesToDelete
     *            Kanten die gelöscht werden sollen.
     */
    private void deleteAllEdgesFrom(Edges edgesToDelete) {
        for (Edge edge : edgesToDelete) {
            graph.removeEdge(edge);
        }
    }

    /**
     * Ermittelt alle Kanten, die in der Nähe einer Bildschirmkoordinate verlaufen.
     *
     * @param coords
     *            Bildschirmkoordinate.
     * @return Menge mit den Kanten als Map Starteckenindex auf Liste der von dort beginnenden
     *         Kanten, die in der Nähe der Bildschirmkoordinate verlaufen.
     */
    private Edges getAllEdgesAt(PixelPoint coords) {
        Edges edgesAtCoords = new Edges();
        int maxDistance = 20;

        for (Edge edge : graph.getEdges()) {
            double distance = calculateDistanceFromCoordsToEdge(coords, edge);
            if (distance < maxDistance) {
                edgesAtCoords.add(edge);
            }
        }

        return edgesAtCoords;
    }

    private double calculateDistanceFromCoordsToEdge(PixelPoint coords, Edge edge) {
        Vertex startVertex = graph.getVertex(edge.getStartVertexIndex());
        Vertex targetVertex = graph.getVertex(edge.getTargetVertexIndex());
        PixelPoint startPoint = startVertex.getCoordinates();
        PixelPoint targetPoint = targetVertex.getCoordinates();
        Line line = new Line(startPoint.toMathPoint(), targetPoint.toMathPoint());

        Geometry geometry = new Geometry();
        return geometry.calculateDistanceBetweenPointAndLine(coords.toMathPoint(), line);
    }

    private void beep() {
        GuiTools.beep();
    }

    /**
     * Fügt eine neue Ecke in den Graphen ein.
     *
     * @param coords
     *            Koordinaten
     * @param radius
     *            Radius der Ecke
     * @param color
     *            Farbe der Ecke
     */
    private void addVertex(PixelPoint coords, int radius, GedColor color) {
        Vertex newVertex = new Vertex(coords, radius, color);
        graph.addVertex(newVertex);
        setChange();
    }

    /**
     * Fügt eine neue Kante in den Graphen ein.
     *
     * @param vertexFrom
     *            Startecke
     * @param vertexTo
     *            Endecke
     * @param edgeColor
     *            Farbe
     */
    private void addEdge(Vertex vertexFrom, Vertex vertexTo, GedColor edgeColor) {
        int indexFrom = vertexFrom.getIndex();
        int indexTo = vertexTo.getIndex();
        Edge edge = new Edge(indexFrom, indexTo, edgeColor);
        graph.addEdge(edge);
        setChange();
    }

    /**
     * Löst den Kreis von der Bewegung der Maus.
     *
     * @param event
     *            Mousevent.
     */
    private void releaseCircle(MouseEvent event) {
        circleSticksAtMouse = false;
        java.awt.Point point = event.getPoint();
        int mouseX = (int) point.getX();
        int mouseY = (int) point.getY();
        setVertexKoordinates(mouseX, mouseY, stickyVertex);
        /* Ecke Einrasten, falls gewünscht: */
        if (graph.getMeta().isSnapInGrid()) {
            graph.snapVertexToGrid(stickyVertex);
        }
        //System.out.println("release vertex " + stickyVertex);
        stickyVertex = null;
        setChange();
    }

    /**
     * Setzt die Koordinaten der Ecke auf den aktuellen Wert.
     *
     * @param mouseX
     *            x-Koordinate der Maus
     * @param mouseY
     *            y-Koordinate der Maus
     * @param vertex
     *            Zu ändernde Ecke
     */
    private void setVertexKoordinates(int mouseX, int mouseY,
            Vertex vertex) {
        PixelPoint oldCoordinates = vertex.getCoordinates();
        PixelPoint coordinates = new PixelPoint(mouseX, mouseY);
        vertex.setCoordinates(coordinates);
        if (graph.vertexIntersctsWithOtherVertex(vertex)) {
            vertex.setCoordinates(oldCoordinates);
        }
    }

    /**
     * Ermittelt die eindeutige Ecke, die an den gegebenen Koordinaten liegt, falls es eine solche
     * gibt. Anderenfalls (keine in der Nähe oder mehrdeutig) wird null zurück gegeben.
     *
     * @param x
     *            X-Koordinate
     * @param y
     *            Y-Koordinate
     * @return Gefundene eindeutige Ecke oder null
     */
    private Vertex getVertexAt(int x, int y) {
        Vertex meantVertex = null;
        int count = 0;
        for (Vertex vertex : graph.getVertices()) {
            if (hitVertex(x, y, vertex)) {
                meantVertex = vertex;
                ++count;
            }
        }
        if (count == 1) {
            return meantVertex;
        }
        else {
            return null;
        }
    }

    /**
     * Ermittelt die eindeutige Ecke, die an den gegebenen Koordinaten liegt, falls es eine solche
     * gibt. Anderenfalls (keine in der Nähe oder mehrdeutig) wird null zurück gegeben.
     *
     * @param coords
     *            Koordinaten
     * @return Gefundene eindeutige Ecke oder null
     */
    private Vertex getVertexAt(PixelPoint coords) {
        return getVertexAt(coords.getX(), coords.getY());
    }

    /**
     * Ermittelt, ob es Ecken an den gegebenen Koordinaten gibt.
     *
     * @param x
     *            X-Koordinate
     * @param y
     *            Y-Koordinate
     * @return Wahrheitswert
     */
    private boolean areThereVerticesAt(int x, int y) {
        for (Vertex vertex : graph.getVertices()) {
            if (hitVertex(x, y, vertex)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Ermittelt, ob es Ecken an den gegebenen Koordinaten gibt.
     *
     * @param coords
     *            Koordinaten
     * @return Wahrheitswert
     */
    private boolean areThereVerticesAt(PixelPoint coords) {
        return areThereVerticesAt(coords.getX(), coords.getY());
    }

    /**
     * Ermittelt, ob die übergebenen Koordinaten im dargestellten Kreis der
     * übergebenen Ecke liegen.
     *
     * @param mouseX
     *            X-Koordinate des Mausklicks.
     * @param mouseY
     *            Y-Koordinate des Mausklicks.
     * @return Wahrheitswert.
     */
    private boolean hitVertex(int mouseX, int mouseY, Vertex vertex) {
        PixelPoint mouseCoordinate = new PixelPoint(mouseX, mouseY);

        PixelPoint center = vertex.getCoordinates();
        int radius = vertex.getRadius();
        if (Vertex.USE_DEFAULT_RADIUS == radius) {
            radius = graph.getMeta().getDefaultRadius();
        }

        if (radius < 15) {
            radius = 15; // Damit man kleine Ecken leichter mit der Maus erwischt
        }
        Circle vertexCircle = new Circle(center.toMathPoint(), radius);

        return vertexCircle.incloses(mouseCoordinate.toMathPoint());
    }

    /** Setzt den Graphen, der dargestellt wird. */
    public void setGraph(Graph graph) {
        //setChange(); nein sonst geht redo und undo nicht
        // Nach laden selbst aufrufen! setNewGraph hier bauen?
        this.graph = graph;
        init();
    }

    /** Getter für den dargestellten Graphen. */
    public Graph getGraph() {
        return graph;
    }

    /** Setzt den Bearbeitungsmodus. */
    public void setModus(EditModus modus) {
        this.editModus = modus;
        int cursorTyp = modus.getMouseCursor();
        this.setCursor(new Cursor(cursorTyp));
    }

    /** Getter für den Bearbeitungsmodus. */
    public EditModus getModus() {
        return editModus;
    }

    /** Änderung mitteilen. */
    private void setChange() {
        notifyObservers();
        repaint();
    }

    /* Für die Beobachter, die die Änderungen mitbekommen wollen: */
    private List<Observer> observers = new ArrayList<Observer>();

    @Override
    public void addObserver(Observer observer) {
        observers.add(observer);
    }

    @Override
    public void deleteObserver(Observer observer) {
        observers.remove(observer);
    }

    @Override
    public void notifyObservers() {
        for (Observer observer : observers) {
            observer.update();
        }
    }

}
