package de.duehl.math.graph.ged.graph.io;

import static de.duehl.math.graph.ged.graph.io.Ged4LoadAndSaveConstants.*;

import de.duehl.math.graph.ged.graph.Graph;
import de.duehl.math.graph.ged.graph.MetaData;
import de.duehl.math.graph.ged.graph.edge.AdjacenceEntry;
import de.duehl.math.graph.ged.graph.edge.AdjacenceList;
import de.duehl.math.graph.ged.graph.edge.AdjacenceLists;
import de.duehl.math.graph.ged.graph.vertex.Vertex;
import de.duehl.math.graph.ged.graph.vertex.Vertices;

/**
 * Diese Klasse erzeugt eine Darstellung des Graphen in ged4-Format zum Speichern in einer Datei.
 *
 * @version 1.01     2019-05-21
 * @author Christian Dühl
 */

public class Ged4FormatSaver {

    /** Eckenmenge des Graphen. */
    private final Vertices vertices;

    /** Kantenmenge des Graphen in Form von Adjazenzlisten. */
    private final AdjacenceLists adjacenceLists;

    /** Enthält weitere Informationen zum Graphen. */
    private MetaData meta;

    /** In dem Builder wird der zu sichernde String aufgebaut. */
    private StringBuilder builder;

    /**
     * Konstruktor.
     *
     * @param graph
     *            Der zu speichernde Graph.
     */
    public Ged4FormatSaver(Graph graph) {
        this.meta = graph.getMeta();
        this.vertices = graph.getVertices();
        this.adjacenceLists = graph.getAdjacenceLists();
    }

    /** Gibt den Graphen als String gespeichert im ged3-Format zurück. */
    public String createSaveString() {
        builder = new StringBuilder();

        appendMetaData();
        appendVertices();
        appendEdges();

        return builder.toString();
    }

    /** Metadaten hinzufügen. */
    private void appendMetaData() {
        builder.append(META_LINE_START + ";");
        builder.append(META_DEFAULT_VERTEX_COLOR + "="
                + meta.getDefaultVertexColor().toColorString() + ";");
        builder.append(
                META_DEFAULT_EDGE_COLOR + "=" + meta.getDefaultEdgeColor().toColorString() + ";");
        builder.append(META_DEFAULT_RADIUS + "=" + meta.getDefaultRadius() + ";");
        builder.append(
                META_BACKGROUND_COLOR + "=" + meta.getBackgroundColor().toColorString() + ";");
        builder.append(META_SHOW_GRID + "=" + meta.isShowGrid() + ";");
        builder.append(META_GRID_DISTANCE + "=" + meta.getGridDistance() + ";");
        builder.append(META_SNAP_IN_GRID + "=" + meta.isSnapInGrid() + ";");
        builder.append(META_GRID_COLOR + "=" + meta.getGridColor().toColorString() + ";");
        builder.append(META_BACKGROUND_PICTURE + "=" + meta.getBackgroundPicture() + ";");
        builder.append(META_DEFAULT_EDGE_WIDTH + "=" + meta.getDefaultEdgeWidth());
        builder.append("\n");
    }

    /** Ecken hinzufügen. */
    private void appendVertices() {
        for (Vertex vertex : vertices) {
            appendVertex(vertex);
        }
    }

    private void appendVertex(Vertex vertex) {
        builder.append(VERTEX_LINE_START + ";");
        builder.append(VERTEX_NUMBER + "=" + vertex.getIndex() + ";");
        builder.append(VERTEX_X + "=" + vertex.getCoordinates().getX() + ";");
        builder.append(VERTEX_Y + "=" + vertex.getCoordinates().getY() + ";");
        builder.append(VERTEX_COLOR + "=" + vertex.getColor().toColorString() + ";");
        builder.append(VERTEX_RADIUS + "=" + vertex.getRadius() + ";");
        String label = hideUsedSignsInLabel(vertex.getLabel());
        builder.append(VERTEX_LABEL + "=" + label + ";");
        builder.append(VERTEX_LABEL_X_DISTANCE + "=" + vertex.getLabelDistance().getX() + ";");
        builder.append(VERTEX_LABEL_Y_DISTANCE + "=" + vertex.getLabelDistance().getY() + ";");
        builder.append(VERTEX_LABEL_COLOR + "=" + vertex.getLabelColor().toColorString() + ";");
        builder.append(VERTEX_WEIGHT + "=" + vertex.getWeight() + ";");
        builder.append(VERTEX_WEIGHT_X_DISTANCE + "=" + vertex.getWeightDistance().getX() + ";");
        builder.append(VERTEX_WEIGHT_Y_DISTANCE + "=" + vertex.getWeightDistance().getY() + ";");
        builder.append(VERTEX_WEIGHT_COLOR + "=" + vertex.getWeightColor().toColorString());
        builder.append("\n");
    }

    private String hideUsedSignsInLabel(String label) {
        String cleanedLabel = label;
        cleanedLabel = cleanedLabel.replace(";", ",");
        cleanedLabel = cleanedLabel.replace("=", "-");
        return cleanedLabel;
    }

    /** Kanten hinzufügen. */
    private void appendEdges() {
        for (int startIndex=0; startIndex < adjacenceLists.size(); ++startIndex) {
            AdjacenceList list = adjacenceLists.get(startIndex);
            appendAdjacenceList(startIndex, list);
        }
    }

    private void appendAdjacenceList(int startIndex, AdjacenceList list) {
        for (AdjacenceEntry entry : list) {
            appendAdjacenceEntry(startIndex, entry);
        }
    }

    /*
     * TODO: Liste von Gleichungen bilden, in Methode die das "=" dazwischen schreibt, dann alles
     * mit Text.join... zusammenpappen
     */

    private void appendAdjacenceEntry(int startIndex, AdjacenceEntry entry) {
        builder.append(EDGE_LINE_START + ";");
        builder.append(EDGE_START_VERTEX + "=" + startIndex + ";");
        builder.append(EDGE_END_VERTEX + "=" + entry.getNeighbourIndex() + ";");
        builder.append(EDGE_COLOR + "=" + entry.getColor().toColorString() + ";");
        builder.append(EDGE_STYLE + "=" + entry.getStyle() + ";");
        builder.append(EDGE_WIDTH + "=" + entry.getWidth() + ";");
        String label = hideUsedSignsInLabel(entry.getLabel());
        builder.append(EDGE_LABEL + "=" + label + ";");
        builder.append(EDGE_LABEL_X_DISTANCE + "=" + entry.getLabelDistance().getX() + ";");
        builder.append(EDGE_LABEL_Y_DISTANCE + "=" + entry.getLabelDistance().getY() + ";");
        builder.append(EDGE_LABEL_COLOR + "=" + entry.getLabelColor().toColorString() + ";");
        builder.append(EDGE_WEIGHT + "=" + entry.getWeight() + ";");
        builder.append(EDGE_WEIGHT_X_DISTANCE + "=" + entry.getWeightDistance().getX() + ";");
        builder.append(EDGE_WEIGHT_Y_DISTANCE + "=" + entry.getWeightDistance().getY() + ";");
        builder.append(EDGE_WEIGHT_COLOR + "=" + entry.getWeightColor().toColorString() + ";");
        builder.append(EDGE_CAPACITY + "=" + entry.getCapacity() + ";");
        builder.append(EDGE_CAPACITY_X_DISTANCE + "=" + entry.getCapacityDistance().getX() + ";");
        builder.append(EDGE_CAPACITY_Y_DISTANCE + "=" + entry.getCapacityDistance().getY() + ";");
        builder.append(EDGE_CAPACITY_COLOR + "=" + entry.getCapacityColor().toColorString() + ";");
        builder.append(EDGE_FLOW + "=" + entry.getFlow() + ";");
        builder.append(EDGE_FLOW_X_DISTANCE + "=" + entry.getFlowDistance().getX() + ";");
        builder.append(EDGE_FLOW_Y_DISTANCE + "=" + entry.getFlowDistance().getY() + ";");
        builder.append(EDGE_FLOW_COLOR + "=" + entry.getFlowColor().toColorString());
        builder.append("\n");
    }

}
