package de.duehl.math.graph.ged.graph.subGraph;

import java.util.ArrayList;
import java.util.List;

import de.duehl.math.graph.ged.graph.Graph;
import de.duehl.math.graph.ged.graph.edge.AdjacenceEntry;
import de.duehl.math.graph.ged.graph.vertex.Vertex;

/**
 * Diese Klasse erstellt Teilgraphen mit einer Ecke und ihren Nachbarn.
 *
 * @version 1.02     2019-05-27
 * @author Christian Dühl
 */

public class VertexWithItsNeigboursSubGraphCreator extends SubgraphCreator {

    /** Die Startecke im originalen Graphen. */
    private Vertex startVertex;

    /**
     * Die Liste der Nachbarn der Startecke im originalen Graphen, zu denen und von denen Kanten
     * von der Startecke aus laufen.
     */
    private List<Vertex> originalNeighbours;

    /** Kopie der Startecke im Teilgraphen. */
    private Vertex newStartVertex;

    /**
     * Menge aller originalen Ecken die wir im Teilgraphen darstellen wollen: Ausgangsecke und die
     * Nachbarn:
     */
    private List<Vertex> interesstingOriginalVertices;

    /**
     * Indices zu der Menge aller originalen Ecken die wir im Teilgraphen darstellen wollen: Ausgangsecke und die
     * Nachbarn:
     */
    private List<Integer> interesstingOriginalVertexIndices;

    /**
     * Konstruktor.
     *
     * @param graph
     *            Graph zu dem Teilgraphen zu erstellen sind.
     * @param subGraphSize
     *            Größe der quadratischen Teilgraphen.
     */
    public VertexWithItsNeigboursSubGraphCreator(Graph graph, int subGraphSize) {
        super(graph, subGraphSize, subGraphSize);
    }

    /**
     * Konstruktor.
     *
     * @param graph
     *            Graph zu dem Teilgraphen zu erstellen sind.
     * @param subGraphSizeX
     *            Breite der Teilgraphen.
     * @param subGraphSizeY
     *            Höhe der Teilgraphen.
     */
    public VertexWithItsNeigboursSubGraphCreator(Graph graph, int subGraphSizeX, int subGraphSizeY) {
        super(graph, subGraphSizeX, subGraphSizeY);
    }

    /**
     * Erzeugt einen Teilgraphen der eine Ecke mit ihren Nachbarn anzeigt.
     *
     * @param startIndex
     *            Index der Startecke.
     */
    public Graph createVertexWithItsNeigboursSubGraph(int startIndex) {
        return createVertexWithItsNeigboursSubGraph(graph.getVertex(startIndex));
    }

    /**
     * Erzeugt einen Teilgraphen der eine Ecke mit ihren Nachbarn anzeigt.
     *
     * @param startVertex
     *            Startecke.
     */
    public Graph createVertexWithItsNeigboursSubGraph(Vertex startVertex) {
        this.startVertex = startVertex;

        collectAllNeighbours();

        createSubGraphWithCopiedMeta();

        createCopyOfStartVertexForSubGraph();
        addVerticesToSubGraph();

        createListOfInteresstingOriginalVertices();
        createInteresstingOriginalVertexIndices();

        addNewEdgesToSubGraph();

        //maximizeSubGraph(subGraph);
        maximizeSubGraph(subGraph, newStartVertex);

        return subGraph;
    }

    /* erstellt die originalen Indices dieser Ecken. */
    private void createInteresstingOriginalVertexIndices() {
        interesstingOriginalVertexIndices = new ArrayList<>();
        for (Vertex vertex : interesstingOriginalVertices) {
            interesstingOriginalVertexIndices.add(vertex.getIndex());
        }
    }

    /** Sammelt alle Nachbarn zu denen und von denen Kanten von der Startecke aus laufen. */
    private void collectAllNeighbours() {
        originalNeighbours = graph.getForwardsAndBackwardsNeighboursAsVertices(startVertex);
    }

    private void createCopyOfStartVertexForSubGraph() {
        newStartVertex = new Vertex(startVertex);
    }

    /** Fügt Kopien der interessanten Ecken zum neuen Teilgraphen hinzu. */
    private void addVerticesToSubGraph() {
        subGraph.addVertex(newStartVertex); // Index 0 - wird automatisch korrigiert
        for (Vertex neighbourVertex : originalNeighbours) {
            Vertex subGraphVertex = new Vertex(neighbourVertex);
            subGraph.addVertex(subGraphVertex); // Index 1 bis n - wird automatisch korrigiert.
        }
    }

    /** Erstellt die Menge aller originaler Ecken: Ausgangsecke und die Nachbarn. */
    private void createListOfInteresstingOriginalVertices() {
        interesstingOriginalVertices = new ArrayList<>();
        interesstingOriginalVertices.add(startVertex);
        interesstingOriginalVertices.addAll(originalNeighbours);

        if (subGraph.getNumberOfVertices() != interesstingOriginalVertices.size()) {
            throw new RuntimeException("Denkfehler...");
        }
        if (!interesstingOriginalVertices.get(0).equals(startVertex)) {
            throw new RuntimeException("Denkfehler...");
        }
        if (!interesstingOriginalVertices.get(0).equalsIgnoreIndex(newStartVertex)) {
            throw new RuntimeException("Denkfehler...");
        }

        for (int index = 0; index < originalNeighbours.size(); ++index) {
            Vertex vertex = originalNeighbours.get(index);
            if (!interesstingOriginalVertices.get(index + 1).equals(vertex)) {
                throw new RuntimeException("Denkfehler...");
            }
        }
    }

    /* Neue Kanten erzeugen: */
    private void addNewEdgesToSubGraph() {
        for (Vertex originalVertex : interesstingOriginalVertices) {
            addNewEdgesToSubGraph(originalVertex);
        }
    }

    private void addNewEdgesToSubGraph(Vertex originalVertex) {
        int indexOfOriginalVertex = originalVertex.getIndex();
        int indexOfSubGraphVertex = originalIndexToSubGraphIndex(indexOfOriginalVertex);

        for (AdjacenceEntry originalEntry : graph.getAdjacenceList(originalVertex)) {
            int originalNeighbourIndex = originalEntry.getNeighbourIndex();
            if (interesstingOriginalVertexIndices.contains(originalNeighbourIndex)) {
                AdjacenceEntry subGraphEntry = new AdjacenceEntry(originalEntry);

                int subGraphNeighbourIndex = originalIndexToSubGraphIndex(originalNeighbourIndex);
                subGraphEntry.setNeighbourIndex(subGraphNeighbourIndex);

                subGraph.addEdge(indexOfSubGraphVertex, subGraphEntry);
            }
        }
    }

    /*
     * Erklärung zur Zielindex-Berechnung:
     *
     * Der Index des Zieles der Kante bestimmt sich durch die Reihenfolge der Nachbarn, denn genau
     * in dieser wurden die Ecken in den neuen Graphen eingefügt.
     *
     * Der Index muss für die Kanten vor dem Einfügen stimmen. Anders als bei den Ecken, da wird er
     * automatisch korrigiert.
     */
    private int originalIndexToSubGraphIndex(int originalIndex) {
        return interesstingOriginalVertexIndices.indexOf(originalIndex);
    }

}
