package de.duehl.math.graph.ged.logic;

import java.util.Map;

import de.duehl.math.graph.ged.graph.Graph;
import de.duehl.math.graph.ged.graph.edge.AdjacenceEntry;
import de.duehl.math.graph.ged.graph.edge.AdjacenceList;
import de.duehl.math.graph.ged.graph.vertex.Vertex;
import de.duehl.math.graph.ged.pixel.PixelPoint;
import de.duehl.basics.logging.Logger;
import de.duehl.basics.logic.Version;

/**
 * Dieses Interface steht für die Schnittstelle zur Logik des Programms.
 *
 * @version 1.01     2019-05-27
 * @author Christian Dühl
 */

public interface Logic {

    /** Erzeugt einen neuen, leeren Graphen. */
    void createEmptyFile();

    /** Beendet das Programm. */
    void quitProgram();

    /** Setzt den anzuzeigenden Status. */
    void setStatus(String string);

    /**
     * Fragt nach dem zu ladenden Dateinamen und lädt diesen neuen Graphen aus
     * dem .ged Format.
     */
    void loadGedFile();

    /**
     * Fragt nach dem zu ladenden Dateinamen und lädt diesen neuen Graphen aus
     * dem .ged2-Format.
     */
    void loadGed2File();

    /**
     * Fragt nach dem zu ladenden Dateinamen und lädt diesen neuen Graphen aus
     * dem .ged3-Format.
     */
    void loadGed3File();

    /**
     * Fragt nach dem zu ladenden Dateinamen und lädt diesen neuen Graphen aus
     * dem .ged4-Format.
     */
    void loadGed4File();

    /** Unterbricht eine ggf. begonnene Bearbeitung. */
    void interruptEditing();

    /** Speichert den Graphen im .ged-Format. */
    void saveGedFile();

    /** Speichert den Graphen im .ged2-Format. */
    void saveGed2File();

    /** Speichert den Graphen im .ged3-Format. */
    void saveGed3File();

    /** Speichert den Graphen im .ged4-Format. */
    void saveGed4File();

    /**
     * Speichert den Graphen (im .ged4-Format) unter einem anderen Namen, nach
     * dem der Benutzer gefragt wird.
     */
    void saveGed4FileAs();

    /** Einen Schritt in der Änderunghistorie zurück. */
    void undo();

    /** Einen Schritt in der Änderunghistorie vorwärts. */
    void redo();

    /** Speichert die aktuelle Darstellung des Graphen im Panel als Bild ab. */
    void screenhot();

    /** Wählt die Defaultfarbe für die Ecken aus. */
    void chooseDefaultVertexColor();

    /** Wählt die Defaultfarbe für die Kanten aus. */
    void chooseDefaultEdgeColor();

    /** Zeigt den Dialog zum Bearbeiten der Eckeneigenschaften an. */
    void showVertexPopupMenu(Vertex vertex, PixelPoint coords);

    /** Zeigt den Dialog zum Bearbeiten der Kanteneigenschaften an. */
    void showEdgePopupMenu(int theIndex, AdjacenceEntry theEntry, PixelPoint coords);

    /** Zeigt den Graphen in Textform an. */
    void showGraph();

    /**
     * Bestimmt aus einer Menge von Kanten, welche davon gelöscht werden sollen
     *
     * @param map
     *            Menge mit den Kanten als Map Starteckenindex auf Liste der von
     *            dort beginnenden Kanten, die in der Nähe einer bestimmten
     *            Bildschirmkoordinate verlaufen.
     * @param coords
     *            Koordinaten
     * @return Ebensolches Map, das nur noch die Kanten enthält, die entfernt
     *         werden sollen.
     */
    Map<Integer, AdjacenceList> determineEdgesToDelete(Map<Integer, AdjacenceList> map,
            PixelPoint coords);

    /** Schaltet das Anzeigen des Gitters um. */
    void toggleMetaShowGrid();

    /** Schaltet das Einrasten an den Gitterlinien ein oder aus. */
    void toggleMetaSnapInGrid();

    /** Wählt den Defaultradius für Ecken. */
    void chooseDefaultVertexRadius();

    /** Wählt die Standarddicke für Kanten. */
    void chooseDefaultEdgeWidth();

    /** Wählt den Abstand der Gitterlinien. */
    void chooseGridDistance();

    /** Wählt die Gitterfarbe aus. */
    void chooseGridColor();

    /** Wählt die Hintergrundfarbe aus. */
    void chooseBackgroundColor();

    /** Wählt ein Hintergrundbild aus. */
    void chooseBackgroundPicture();

    /** Entfernt das Hintergrundbild. */
    void removeBackgroundPicture();

    /** Bewegt den Graphen um den Gitterabstand nach links. */
    void moveLeft();

    /** Bewegt den Graphen um den Gitterabstand nach rechts. */
    void moveRight();

    /** Bewegt den Graphen um den Gitterabstand nach oben. */
    void moveUp();

    /** Bewegt den Graphen um den Gitterabstand nach unten. */
    void moveDown();

    /**
     * Zentriert den Graphen auf den Mittelpunkt der sichtbaren Fläche.
     *
     * @param minX
     *            x-Koordinate der linken oberen Ecke der Fläche.
     * @param minY
     *            y-Koordinate der linken oberen Ecke der Fläche.
     * @param maxX
     *            x-Koordinate der rechten unteren Ecke der Fläche.
     * @param maxY
     *            y-Koordinate der rechten unteren Ecke der Fläche.
     */
    void centerGraph(int minX, int minY, int maxX, int maxY);

    /**
     * Zentriert den Graphen auf den Mittelpunkt der sichtbaren Fläche und
     * vergrößert ihn so weit er dann noch auf den Schirm passt.
     *
     * @param minX
     *            x-Koordinate der linken oberen Ecke der Fläche.
     * @param minY
     *            y-Koordinate der linken oberen Ecke der Fläche.
     * @param maxX
     *            x-Koordinate der rechten unteren Ecke der Fläche.
     * @param maxY
     *            y-Koordinate der rechten unteren Ecke der Fläche.
     */
    void maximizeGraph(int minX, int minY, int maxX, int maxY);

    /** Verkleinert den Graphen.
     *
     * @param minX
     *            x-Koordinate der linken oberen Ecke der Fläche.
     * @param minY
     *            y-Koordinate der linken oberen Ecke der Fläche.
     * @param maxX
     *            x-Koordinate der rechten unteren Ecke der Fläche.
     * @param maxY
     *            y-Koordinate der rechten unteren Ecke der Fläche.
     */
    void reduceGraph(int minX, int minY, int maxX, int maxY);

    /** Vergrößert den Graphen.
     *
     * @param minX
     *            x-Koordinate der linken oberen Ecke der Fläche.
     * @param minY
     *            y-Koordinate der linken oberen Ecke der Fläche.
     * @param maxX
     *            x-Koordinate der rechten unteren Ecke der Fläche.
     * @param maxY
     *            y-Koordinate der rechten unteren Ecke der Fläche.
     */
    void enlargeGraph(int minX, int minY, int maxX, int maxY);

    /** Lässt alle Ecken des Graphen im Gitter einrasten. */
    void snapAllVerticesToGrid();

    /** Zeigt die Größe des GraphPanels an. */
    void showGraphPanelSize(int width, int height);

    /** Setzt eine Ecke in den Mittelpunkt der Bildfläche. */
    void insertCenter();

    /** Fügt einen Kreis um den Mittelpunkt der aktuellen Bildfläche ein. */
    void insertCircle(int n);

    /**
     * Fügt einen vollständigen Graphen um den Mittelpunkt der aktuellen
     * Bildfläche ein.
     */
    void insertCompleteGraph(int n);

    /** Zeigt die Webseite an. */
    void webpage();

    /**
     * Setzt die Größe der Canvas.
     *
     * @param width
     *            neue Breite
     * @param height
     *            neue Höhe
     */
    void setViewSize(int i, int j);

    /** Gibt den Graphen zurück, der angezeigt wird. */
    Graph getGraph();

    /** Gibt den Logger zurück. */
    Logger getLogger();

    /** Setzt die Farben aller Ecken auf Default. */
    void setAllVertexColorsToDefault();

    /** Setzt die Radien aller Ecken auf Default. */
    void setAllVertexRadiusToDefault();

    /** Setzt die Farben aller Kanten auf Default. */
    void setAllEdgeColorsToDefault();

    /** Setzt die Dicken aller Kanten auf Default. */
    void setAllEdgeWidthsToDefault();

    /** Getter für das Verzeichnis ~/.ged für Einstellungen und Graphendateien. */
    String getDirectory();

    /** Getter für die Programmversion. */
    Version getVersion();

    /** Zum Mitteilen von Änderungen am Graphen. */
    void update();

}
