package de.duehl.math.graph.ged.ui.elements.color;

import java.awt.BorderLayout;
import java.awt.Color;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;

import de.duehl.math.graph.ged.graph.GedColor;
import de.duehl.math.graph.ged.ui.GedGuiTools;
import de.duehl.swing.ui.colors.ColorTool;

/**
 * Diese Klasse stellt für den Ecken- und Kanten-Dialog eine Farbwahl mit Default-Schalter dar.
 *
 * @version 1.01     2019-05-24
 * @author Christian Dühl
 */

public class DefaultOrOwnColorChooser {

    /** Gibt an, ob zu Beginn die Default-Farbe verwendet wird. */
    private final boolean hasDefaultColor;

    /** Text der auf der Checkbox zum an und Ausschalten des Default-Zustands angezeigt wird. */
    private final String textForToggleButton;

    /** Text der als Titel vor dem Farbbutton angezeigt wird. */
    private final String textForColorButtonTitle;

    /** Text der als Titel des Dialogs zur Farbauswahl angezeigt wird. */
    private final String textForColorChooserDialogTitle;

    /** Farbe des Objekts zum Zeitpunkt des Aufrufs. */
    private final Color actualSwingColor;

    /** Default-Farbe des Objekts. */
    private GedColor defaultGedColor;

    /** Default-Farbe des Objekts. */
    private Color defaultSwingColor;

    /** Setzt die Farbe im Objekt. */
    private final ColorSetter colorSetter;

    /** Rahmen von GED zur relativen Positionierung des Farbauswahldialogs. */
    private final JFrame frame;

    /**
     * Titel vor dem Button zur Farbwahl, dieser wird erst aktiv, wenn nicht mehr die Defaultfarbe
     * verwendet wird.
     */
    private JLabel colorButtonTitleLabel;
    /**
     * Button zur Farbwahl, dieser wird erst aktiv, wenn nicht mehr die Defaultfarbe verwendet
     * wird.
     */
    private JButton colorButton;

    private JCheckBox defaultCheckBox;

    /**
     * Konstruktor.
     *
     * @param hasDefaultColor
     *            Gibt an, ob zu Beginn die Default-Farbe verwendet wird.
     * @param textForToggleButton
     *            Text der auf der Checkbox zum an und Ausschalten des Default-Zustands angezeigt
     *            wird.
     * @param textForColorButtonTitle
     *            Text der als Titel vor dem Farbbutton angezeigt wird.
     * @param textForColorChooserDialogTitle
     *            Text der als Titel des Dialogs zur Farbauswahl angezeigt wird.
     * @param actualSwingColor
     *            Farbe des Objekts zum Zeitpunkt des Aufrufs.
     * @param defaultGedColor
     *            Default-Farbe des Objekts.
     * @param colorSetter
     *            Setzt die Farbe im Objekt.
     * @param frame
     *            Rahmen von GED zur relativen Positionierung des Farbauswahldialogs.
     */
    public DefaultOrOwnColorChooser(boolean hasDefaultColor, String textForToggleButton,
            String textForColorButtonTitle, String textForColorChooserDialogTitle,
            Color actualSwingColor, GedColor defaultGedColor, ColorSetter colorSetter,
            JFrame frame) {
        this.hasDefaultColor = hasDefaultColor;
        this.textForToggleButton = textForToggleButton;
        this.textForColorButtonTitle = textForColorButtonTitle;
        this.textForColorChooserDialogTitle = textForColorChooserDialogTitle;
        this.actualSwingColor = actualSwingColor;
        this.defaultGedColor = defaultGedColor;
        this.defaultSwingColor = defaultGedColor.toSwingColor();
        this.colorSetter = colorSetter;
        this.frame = frame;
    }

    public JPanel createPanel() {
        JPanel panel = new JPanel();
        panel.setLayout(GedGuiTools.createVerticalComponentLayout());

        panel.add(createUseDefaultColorCheckBox());
        panel.add(createColorPanel());

        return panel;
    }

    private JComponent createUseDefaultColorCheckBox() {
        defaultCheckBox = new JCheckBox(textForToggleButton);
        defaultCheckBox.setSelected(hasDefaultColor);
        defaultCheckBox.addActionListener(e -> useDefaultColor());
        return defaultCheckBox;
    }

    private void useDefaultColor() {
        boolean isDefaultChecboxSelected = defaultCheckBox.isSelected();

        if (isDefaultChecboxSelected) {
            colorSetter.setColor(GedColor.DEFAULT); // -1 -1 -1 Hinweis auf Defaultnutzung
        }
        else {
            colorSetter.setColor(defaultGedColor); // echte Defaultfarbe zur Initialisierung
        }

        colorButtonTitleLabel.setEnabled(!isDefaultChecboxSelected);
        colorButton.setEnabled(!isDefaultChecboxSelected);
        setColorButtonColor(defaultSwingColor);
    }

    private JComponent createColorPanel() {
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());

        panel.add(createColorLabel(), BorderLayout.WEST);
        panel.add(createColorButton(), BorderLayout.CENTER);

        return panel;
    }

    private JComponent createColorLabel() {
        colorButtonTitleLabel = new JLabel(textForColorButtonTitle);
        return colorButtonTitleLabel;
    }

    private JComponent createColorButton() {
        colorButton = new JButton("   ändern   ");

        Color bgc = hasDefaultColor ? defaultSwingColor : actualSwingColor;
        Color fgc = new ColorTool().anticolor(bgc);

        colorButton.setBackground(bgc);
        colorButton.setForeground(fgc);
        colorButton.addActionListener(e -> selectColor());
        colorButton.setEnabled(!hasDefaultColor);

        return colorButton;
    }

    private void selectColor() {
        // GedColor oldColor = entry.getColor();
        Color swingColor = JColorChooser.showDialog(frame, textForColorChooserDialogTitle,
                colorButton.getBackground());
        if (null != swingColor) {
            setColorButtonColor(swingColor);
            colorSetter.setColor(new GedColor(swingColor));
        }
    }

    private void setColorButtonColor(Color swingColor) {
        colorButton.setForeground(new ColorTool().anticolor(swingColor));
        colorButton.setBackground(swingColor);
    }

    public void setEnabled(boolean enabled) {
        defaultCheckBox.setEnabled(enabled);

        boolean enableLowerColorPart = !defaultCheckBox.isSelected() && enabled;
        colorButtonTitleLabel.setEnabled(enableLowerColorPart);
        colorButton.setEnabled(enableLowerColorPart);
    }

    public boolean isDefaultColorSelected() {
        return defaultCheckBox.isSelected();
    }

    public GedColor getSelectedColor() {
        return new GedColor(colorButton.getBackground());
    }

    public void setDefaultButtonColor(GedColor gedColor) {
        defaultGedColor = gedColor;
        defaultSwingColor = defaultGedColor.toSwingColor();

        if (defaultCheckBox.isSelected()) {
            setColorButtonColor(defaultSwingColor);
        }
    }

}
