package de.duehl.math.graph.ged.ui.elements.radiusandwidth;

import java.awt.BorderLayout;
import java.awt.Component;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPanel;

import de.duehl.math.graph.ged.pixel.PixelPoint;
import de.duehl.math.graph.ged.ui.GedGuiTools;
import de.duehl.math.graph.ged.ui.dialogs.SizeChooser;
import de.duehl.swing.ui.components.elements.HorizontalNumericTextFieldWithTitle;
import de.duehl.swing.ui.components.elements.TextFieldWithTitle;

/**
 * Diese Klasse stellt ein Element zur Auswahl eines Integerwertes etwa für den Radius einer Ecke
 * oder die Dicke einer Kante dar.
 *
 * @version 1.01     2019-05-27
 * @author Christian Dühl
 */

public class RadiusOrWidthChooser {

    /** Gibt an, ob zu Beginn der Default-Wert verwendet wird. */
    private final boolean hasDefaultValue;

    /** Der Text der CheckBox zur Verwendung des Default-Wertes. */
    private final String defaultCheckBoxText;

    /** Der Wert der anzeigt, dass der Default verwendet wird. */
    private final int defaultValue;

    /** Der in den Metadaten als Defaultwert hinterlegte Wert. */
    private final int metaDefaultValue;

    /** Titel vor dem Eingabefeld für einen individuellen Wert. */
    private final String textFieldTitle;

    /** Setzt Radius der Ecke oder Dicke der Kante im Objekt. */
    private final RadiusOrWidthSetter valueSetter;

    /** Koordinaten, an denen der Dialog angezeigt werden soll. */
    private final PixelPoint coords;


    private final JPanel panel;

    /** Das Feld zur Eingabe der Dicke der Kante. */
    private TextFieldWithTitle valueField;

    private JCheckBox defaultCheckBox;

    /** Der Button zur Auswahl einer bestimmten Dicke. */
    private JButton valueChooseButton;

    /**
     * Konstruktor.
     *
     * @param hasDefaultValue
     *            Gibt an, ob zu Beginn der Default-Wert verwendet wird.
     * @param defaultCheckBoxText
     *            Der Text der CheckBox zur Verwendung des Default-Wertes.
     * @param defaultValue
     *            Der Wert der anzeigt, dass der Default verwendet wird.
     * @param metaDefaultValue
     *            Der in den Metadaten als Defaultwert hinterlegte Wert.
     * @param textFieldTitle
     *            Titel vor dem Eingabefeld für einen individuellen Wert.
     * @param valueSetter
     *            Setzt Radius der Ecke oder Dicke der Kante im Objekt.
     * @param coords
     *            Koordinaten, an denen der Dialog angezeigt werden soll.
     */
    public RadiusOrWidthChooser(boolean hasDefaultValue, String defaultCheckBoxText,
            int defaultValue, int metaDefaultValue, String textFieldTitle,
            RadiusOrWidthSetter valueSetter, PixelPoint coords) {
        this.hasDefaultValue = hasDefaultValue;
        this.defaultCheckBoxText = defaultCheckBoxText;
        this.defaultValue = defaultValue;
        this.metaDefaultValue = metaDefaultValue;
        this.textFieldTitle = textFieldTitle;
        this.valueSetter = valueSetter;
        this.coords = coords;
        panel = createPanel();
    }

    /** Erzeugt den Panel mit den Elementen zur Dicke der Kante. */
    private JPanel createPanel() {
        JPanel panel = new JPanel();
        panel.setLayout(GedGuiTools.createVerticalComponentLayout());

        panel.add(createDefaultCheckBox());
        panel.add(createWidthPanel());

        return panel;
    }

    private Component createDefaultCheckBox() {
        defaultCheckBox = new JCheckBox(defaultCheckBoxText);
        defaultCheckBox.setSelected(hasDefaultValue);
        defaultCheckBox.addActionListener(e -> useDefault(defaultCheckBox));
        return defaultCheckBox;
    }

    private void useDefault(JCheckBox checkBox) {
        if (checkBox.isSelected()) {
            valueSetter.setRadiusOrWidth(metaDefaultValue); // defaultValue
            valueField.setEnabled(false);
            valueChooseButton.setEnabled(false);
        }
        else {
            valueSetter.setRadiusOrWidth(metaDefaultValue);
            valueField.setEnabled(true);
            valueChooseButton.setEnabled(true);
        }
    }

    private Component createWidthPanel() {
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());

        panel.add(createValueField(), BorderLayout.CENTER);
        panel.add(createWidthChooseButton(), BorderLayout.EAST);

        return panel;
    }

    private Component createValueField() {
        valueField = new HorizontalNumericTextFieldWithTitle(textFieldTitle);
        valueField.setEnabled(!hasDefaultValue);
        return valueField.getPanel();
    }

    public void setValue(int value) {
        valueField.setText(Integer.toString(determineWidthText(value)));
    }

    private int determineWidthText(int value) {
        if (value == defaultValue) {
           return metaDefaultValue;
        }
        else {
            return value;
        }
    }

    private Component createWidthChooseButton() {
        valueChooseButton = new JButton("Auswahl");
        valueChooseButton.addActionListener(e -> chooseValue());
        valueChooseButton.setEnabled(!hasDefaultValue);
        return valueChooseButton;
    }

    private void chooseValue() {
        SizeChooser dialog = new SizeChooser();
        int size = dialog.showDialog(coords);
        // TODO: aktuelle Koordinaten statt dessen!
        if (size > 0) {
            valueField.setText(Integer.toString(size));
            valueSetter.setRadiusOrWidth(size);
        }
    }

    public JPanel getPanel() {
        return panel;
    }

    public boolean isDefaultSelected() {
        return defaultCheckBox.isSelected();
    }

    public String getValueText() {
        return valueField.getText();
    }

    public void addChangeAction(Runnable runnable) {
        valueField.addKeyReleasedAction(runnable);
        valueField.addActionListener(e -> runnable.run());
    }

}
