package de.duehl.math.geometry;

/*
 * Copyright 2016 Christian Dühl. All rights reserved.
 *
 * This program is free software. You can redistribute it and/or
 * modify it under the same terms as perl:
 *
 * general:  http://dev.perl.org/licenses/
 * GPL:      http://dev.perl.org/licenses/gpl1.html
 * artistic: http://dev.perl.org/licenses/artistic.html
 */

/**
 * Diese Klasse bietet ein paar geometrische Hilfsfunktionen.
 *
 * @version 1.01     2012-10-30
 * @author Christian Dühl
 */

public class Geometry {

    /** Epsilon für double-Vergleiche. */
    public static final double EPSILON = 0.0001;

    /**
     * Punkt auf einer Linie mit gegebenem Abstand zum Endpunkt der Linie
     * bestimmen.
     *
     * @param line
     *            Linie
     * @param distanceFromEndPoint
     *            Abstand zum Endpunkt der Linie.
     * @return Punkt mit dem gewünschten Abstand oder der Endpunkt, falls der
     *         gegebene Abstand oder der Abstand von Start- und Endpunkt kleiner
     *         als 1.0 ist.
     */
    public Point getPointOnLineWithDistanceToEndPoint(Line line,
            double distanceFromEndPoint) {
        /* Punkte der Linie: */
        Point start = line.getStart();
        Point end   = line.getEnd();

        /*
         * Falls der Abstand vom Endpunkt zu klein ist, einfach diesen zurück
         * geben:
         */
        if (distanceFromEndPoint < 1.0) {
            return end;
        }

        /* Abstand der beiden Punkte berechnen: */
        double distanceX = line.getXDistance();
        double distanceY = line.getYDistance();
        double distance = line.getDistance();

        /*
         * Liegen Start- und Endpunkt zu dicht beieinander, Endpunkt ausgeben,
         * auch wenn der Abstand nicht stimmt.
         */
        if (distance < 1.0) {
            return end;
        }

        /* Koordinaten der Punkte: */
        double startX = start.getX();
        double startY = start.getY();
        double endX = end.getX();
        double endY = end.getY();

        /*
         * Abweichungen berechnen vom Eckenmittelpunkt bis zum
         * Radius des die Ecke darstellenden Kreises:
         */
        double endDx = distanceX * distanceFromEndPoint / distance;
        double endDy = distanceY * distanceFromEndPoint / distance;

        /* Richtiges Vorzeichen für Addition oder Subtraktion: */
        if (startX < endX) {
            endDx = -endDx;
        }
        if (startY < endY) {
            endDy = -endDy;
        }

        /*
         * Koordinaten reduzieren auf Teil außerhalb der
         * dargestellten Ecken:
         */
        double realEndX = endX + endDx;
        double realEndY = endY + endDy;

        Point realEnd = new Point(realEndX, realEndY);

        return realEnd;
    }

    /**
     * Punkt auf einer Linie mit gegebenem Abstand zum Startpunkt der Linie
     * bestimmen.
     *
     * @param line
     *            Linie
     * @param distanceFromStartPoint
     *            Abstand zum Startpunkt der Linie.
     * @return Punkt mit dem gewünschten Abstand oder der Startpunkt, falls der
     *         gegebene Abstand oder der Abstand von Start- und Endpunkt kleiner
     *         als 1.0 ist.
     */
    public Point getPointOnLineWithDistanceToStartPoint(Line line,
            double distanceFromStartPoint) {
        Line swappedLine = new Line(line);
        swappedLine.swap();

        return getPointOnLineWithDistanceToEndPoint(swappedLine,
                distanceFromStartPoint);
    }

    /**
     * Schnittpunkt einer Linie mit einem Kreis bestimmen, wobei die Linie im
     * Mittelpunkt des Kreises endet.
     *
     * @param line
     *            Linie
     * @param circle
     *            Kreis.
     * @return Schnittpunkt der Linie mit dem Kreis.
     */
    public Point getPointOnCircleWhereLineEnds(Line line, Circle circle)
            throws IllegalArgumentException {
        Point end = line.getEnd();
        Point center = circle.getCenter();
        if (!end.equals(center)) {
            throw new IllegalArgumentException("Endpunkt der Linie " + end
                    + " passt nicht zum Mittelpunkt des Kreises " + center
                    + "!");
        }
        return getPointOnLineWithDistanceToEndPoint(line, circle.getRadius());
    }

    /**
     * Schnittpunkt einer Linie mit einem Kreis bestimmen, wobei die Linie im
     * Mittelpunkt des Kreises beginnt.
     *
     * @param line
     *            Linie
     * @param circle
     *            Kreis.
     * @return Schnittpunkt der Linie mit dem Kreis.
     */
    public Point getPointOnCircleWhereLineStarts(Line line, Circle circle)
            throws IllegalArgumentException {
        Point start = line.getStart();
        Point center = circle.getCenter();
        if (!start.equals(center)) {
            throw new IllegalArgumentException("Endpunkt der Linie " + start
                    + " passt nicht zum Mittelpunkt des Kreises " + center
                    + "!");
        }
        return getPointOnLineWithDistanceToStartPoint(line, circle.getRadius());
    }

    /**
     * Berechnet den Abstand des Punktes zur Strecke unter Berücksichtigung der
     * Genauigkeit EPSILON.
     *
     * Vorgehensweise: L = der eine Endpunkt der Linie,
     *                 R = der andere Endpunkt der Linie,
     *                 M = Mittelpunkt,
     *                 schlechtesten Punkt rauswerfen, Endpunkt verschieben,
     *                 dann neuen Mittelpunkt berechnen.
     *
     * @param point
     *            Punkt
     * @param line
     *            Linie
     * @return Abstand
     */
    public double calculateDistanceBetweenPointAndLine(Point point, Line line) {
        Point l = line.getStart();
        Point r = line.getEnd();

        while (calculateDistanceBetweenPointAndPoint(l, r) > EPSILON) {
            /* Mittelpunkt (neu) berechnen: */
            Point m = calculateCenter(r, l);

            /*
             * Falls l oder r nahe genug am Punkt liegen, Abbrechen: (bei Punkt
             * auf Linie Fällen):
             */
            if (point.equals(l, EPSILON) || point.equals(r, EPSILON)) {
                return 0.0;
            }

            /* Abstände des Punktes P zu L, M und R berechnen: */
            double distPtoL = calculateDistanceBetweenPointAndPoint(point, l);
            double distPtoM = calculateDistanceBetweenPointAndPoint(point, m);
            double distPtoR = calculateDistanceBetweenPointAndPoint(point, r);

            /* Falls l oder r nahe genug an m liegen, Abbrechen: */
            if (m.equals(l, EPSILON) || m.equals(r, EPSILON)) {
                return distPtoM;
            }

            /*
             * Falls die Abstände des Punktes zu l und r identisch sind, den
             * Abstand zu m zurückgeben:
             */
            if (doubleEquals(distPtoL, distPtoR)) {
                return distPtoM;
            }

            /* Den schlechteren der Punkte l und r gleich m setzen: */
            if (distPtoL < distPtoR) {
                r = m; // new Point(m); nicht nötig, da m neu erstellt wird.
            }
            else {
                l = m; // new Point(m); nicht nötig, da m neu erstellt wird.
            }
        }

        throw new RuntimeException("Interner Fehler bei der Berechnung.");
    }

    /**
     * Gleichheit für double Werte mit interner Genauigkeit EPSILON.
     *
     * @param d1
     *            Erste zu vergleichende Zahl.
     * @param d2
     *            Zweite zu vergleichende Zahl.
     * @return Wahrheitswert (true: Die Differenz ist kleiner als EPSILON).
     */
    private boolean doubleEquals(double d1, double d2) {
        double diff = Math.abs(d1 - d2);
        return diff < EPSILON;
    }

    /**
     * Berechnet den Abstand zweier Punkte.
     *
     * @param p
     *            Erster Punkt
     * @param q
     *            Zweiter Punkt
     * @return
     */
    public double calculateDistanceBetweenPointAndPoint(Point p, Point q) {
        return calculateDistanceBetweenPointAndPoint(p.getX(), p.getY(),
                q.getX(), q.getY());
    }

    /**
     * Berechnet den Abstand zweier Punkte, die hier als vier Koordinaten
     * übergeben werden.
     *
     * @param px
     *            x-Koordinate vom ersten Punkt (p).
     * @param py
     *            y-Koordinate vom ersten Punkt (p).
     * @param qx
     *            x-Koordinate vom zweiten Punkt (q).
     * @param qy
     *            y-Koordinate vom zweiten Punkt (q).
     * @return
     */
    private double calculateDistanceBetweenPointAndPoint(double px, double py,
            double qx, double qy) {
        double dX = Math.abs(px - qx);
        double dY = Math.abs(py - qy);
        return Math.sqrt(dX * dX + dY * dY);
    }

    /**
     * Berechnet den Mittelpunkt von beliebig vielen Punkten.
     *
     * @param points Punkte
     * @return Mittelpunkt
     */
    public Point calculateCenter(Point... points) {
        double centerX = 0.0;
        double centerY = 0.0;
        for (Point point : points) {
            centerX += point.getX();
            centerY += point.getY();
        }
        double n = points.length;
        centerX /= n;
        centerY /= n;
        return new Point(centerX, centerY);
    }

}
