package de.duehl.math.numbertheory;

/*
 * Copyright 2017 Christian Dühl. All rights reserved.
 *
 * This program is free software. You can redistribute it and/or
 * modify it under the same terms as perl:
 *
 * general:  http://dev.perl.org/licenses/
 * GPL:      http://dev.perl.org/licenses/gpl1.html
 * artistic: http://dev.perl.org/licenses/artistic.html
 */

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.duehl.basics.debug.Assure;
import de.duehl.basics.text.Text;

/**
 * Diese Klasse stellt ein Zahlensystem mit 62 Stellen (0-9, a-z, A-Z) dar, mit dessen Hilfe
 * lange Zahlen aus dem Dezimalsystem platzsparender dargestellt werden können.
 *
 * @version 1.01     2017-09-04
 * @author Christian Dühl
 */

public class SixtytwoNumberingSystem {

    private static final String CHARS = "0123456789"
            + "abcdefghijklmnopqrstuvwxyz"
            + "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    static String getDigitForDecimalValue(int decimal) {
        Assure.isInRange("der übergebene Wert decimal", decimal, 0, CHARS.length() - 1);
        return CHARS.substring(decimal, decimal + 1);
    }

    static long getDecimalValueForDigit(String digit) {
        Assure.isEqual(1, digit.length());
        Assure.isTrue(CHARS.contains(digit));
        return CHARS.indexOf(digit);
    }

    static int getNumberOfDigits() {
        return CHARS.length();
    }

    public static String decimalToSixtytwo(long decimalNumber) {
        Assure.isAtLeast(decimalNumber, 0L);
        if (0 == decimalNumber) {
            return CHARS.substring(0, 1);
        }

        long rest = decimalNumber;
        List<String> digits = new ArrayList<>();
        int numberOfDigits = getNumberOfDigits();

        while (rest > 0) {
            long decimalForOneDigit = rest % numberOfDigits;
            String digit = getDigitForDecimalValue((int) decimalForOneDigit);
            digits.add(digit);
            rest /= numberOfDigits;
        }

        Collections.reverse(digits);
        return Text.join("", digits);
    }

    public static long sixtytwoToDecimal(String sixtyTwoNumber) {
        long decimalNumber = 0;
        int length = sixtyTwoNumber.length();
        int numberOfDigits = getNumberOfDigits();

        for (int index = length - 1; index >= 0; --index) {
            String sixtyTwoDigit = sixtyTwoNumber.substring(index, index + 1);
            int indexFromBehind = length - 1 - index;
            long decimalForDigit = getDecimalValueForDigit(sixtyTwoDigit);
            decimalNumber += decimalForDigit * Math.pow(numberOfDigits, indexFromBehind);
        }

        return decimalNumber;
    }

}
