package de.duehl.swing.ui.tables;

/*
 * Copyright 2017 Christian Dühl. All rights reserved.
 *
 * This program is free software. You can redistribute it and/or
 * modify it under the same terms as perl:
 *
 * general:  http://dev.perl.org/licenses/
 * GPL:      http://dev.perl.org/licenses/gpl1.html
 * artistic: http://dev.perl.org/licenses/artistic.html
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;

import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.table.DefaultTableCellRenderer;

/**
 * Diese Klasse ist für das Rendern (die optische Darstellung) der Tabellenzellen zuständig. Sie
 * wurde von DefaultTableCellRenderer abgeleitet, um nicht für jede Zelle eine neue Klasse zu
 * erzeugen.
 *
 * @version 1.01     2017-03-29
 * @author Christian Dühl
 */

public class DifferentBackgroundsTableRenderer extends DefaultTableCellRenderer {

    private static final long serialVersionUID = 1L;

    public static final float FONT_SIZE = 15f;

    /** Normale Vordergrundfarbe. */
    private final Color lighterForeground;

    /** Normale Hintergrundfarbe. */
    private final Color lighterBackground;

    /** Dunklere Vordergrundfarbe. */
    private final Color darkerForeground;

    /** Dunklere Hintergrundfarbe. */
    private final Color darkerBackground;

    /**
     * Der Konstruktor, hier werden unsere Maps erzeugt.
     *
     * @param foregroundColor
     *            Vordergrundfarbe
     * @param backgroundColor
     *            Hintergrundfarbe
     */
    public DifferentBackgroundsTableRenderer(Color foregroundColor, Color backgroundColor) {
        super();

        this.lighterForeground = calculateLighterColor(foregroundColor);
        this.lighterBackground = calculateLighterColor(backgroundColor);

        this.darkerForeground = calculateDarkerColor(foregroundColor);
        this.darkerBackground = calculateDarkerColor(backgroundColor);
    }

    /**
     * Hier wird aus einer Farbe eine leicht hellere Farbe berechnet.
     *
     * @param color
     *            Farbe.
     * @return hellere Farbe.
     */
    private Color calculateLighterColor(Color color) {
        return changeColor(color, 20);
    }

    /**
     * Hier wird aus einer Farbe eine leicht dunklere Farbe berechnet.
     *
     * @param color
     *            Farbe.
     * @return dunklere Farbe.
     */
    private Color calculateDarkerColor(Color color) {
        return changeColor(color, -20);
    }

    /**
     * Hier wird eine Farbe um den angegebenen Betrag verändert.
     *
     * @param color
     *            Farbe.
     * @return veränderte Farbe.
     */
    private Color changeColor(Color color, int delta) {
        return new Color(
                adjustRangeFrom0To255(color.getRed()   + delta),
                adjustRangeFrom0To255(color.getGreen() + delta),
                adjustRangeFrom0To255(color.getBlue()  + delta));
    }

    private int adjustRangeFrom0To255(int value) {
        if (0 > value) {
            return 0;
        }
        if (255 < value) {
            return 255;
        }
        return value;
    }

    /**
     * Liefert eine Komponente für die anzuzeigende Zelle zurück. Die Hauptarbeit lassen wir
     * super.getTableCellRendererComponent(...) machen und hängen nur ein wenig Code dahinter.
     *
     * @param table
     *            Die Tabelle
     * @param value
     *            Das anzuzeigende Objekt (im Normalfall ein String)
     * @param isSelected
     *            Zeigt an, ob das Objekt selektiert wurde.
     * @param hasFocus
     *            Zeigt an, ob der Focus auf dieser Zelle liegt.
     * @param row
     *            Zeile in der die Zelle liegt.
     * @param column
     *            Spalte in der die Zelle liegt.
     */
    @Override
    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
            boolean hasFocus, int row, int column) {
        JLabel createdLabel = (JLabel) super.getTableCellRendererComponent(table, value, isSelected,
                hasFocus, row, column);
        //createdLabel.setFont(new Font("Arial", Font.BOLD, 30));
        setVerticalAlignment(JLabel.CENTER);
        setHorizontalAlignment(JLabel.LEFT);


        Font oldFont = createdLabel.getFont();
        createdLabel.setFont(oldFont.deriveFont(FONT_SIZE));
        //c.setFont(new Font("Arial", Font.BOLD, 30));
        setVerticalAlignment(JLabel.CENTER);
        setHorizontalAlignment(JLabel.LEFT);

        //table.setRowHeight((int) (tableFontSize * 1.6));
        //table.setRowHeight(20);
        // Das geht nicht, das führt zu rekursiven Aufrufen dieser Methode,
        // wodurch die CPU-Last auf 100% geht.
        // Besser beim erstellen der Tabelle die Zeilenhöhe ändern:
        //     table.setRowHeight((int) (1.6 * tableFontSize));

        boolean odd = row % 2 == 0;
        Color foreground = odd ? lighterForeground : darkerForeground;
        Color background = odd ? lighterBackground : darkerBackground;
        if (isSelected) {
            setForeground(background);
            setBackground(foreground);
        }
        else {
            setForeground(foreground);
            setBackground(background);
        }

        return createdLabel;
    }

}
